!
! A MODULE 
!                               
! "THE BEER-WARE LICENSE":
! Alberto Ramos wrote this file. As long as you retain this 
! notice you can do whatever you want with this stuff. If we meet some 
! day, and you think this stuff is worth it, you can buy me a beer in 
! return. <alberto.ramos@cern.ch>
!
! $ v1.0 201XXXXX $
!                               

submodule (aderrors) aderrors_arith

  implicit none

  type ws_alloc
     integer :: ns = -1, no = -1, nr=-1
     integer, allocatable :: map(:,:), ids(:), nrep(:), &
          ivrep(:), nd(:)
     real (kind=DP), allocatable :: texp(:), s(:)
  end type ws_alloc
  
  type (ws_alloc), save :: ws

contains

! ********************************
! *
  module subroutine addobs_f(x, a, fn)
! *
! ********************************
    type (uwreal), intent (inout) :: x
    type (uwreal), intent (in)    :: a(:)

    integer        :: no, nid, i, j, is_r, ie_r, &
         is_s, ie_s, nvt, isave
    real (kind=DP) :: der(size(a))
    type (uwreal)  :: tfoo
    
    interface 
       function fn(z)
         USE numtypes
         import :: uwreal
         type (uwreal) :: fn
         type (uwreal), intent (in) :: z(:)
       end function fn
    end interface

    no  = size(a)
    nvt = 0
    nid = 0
    do i = 1, no
       nvt = nvt + size(a(i)%ivrep)
       nid = nid + a(i)%nid
    end do
    call init_ws(nid,nvt,no)
    call combine_ids_multi(nid,a)
    nvt = sum(ws%nrep(1:nid))
    call x%init(nid, ws%nd, ws%nrep, ws%ivrep(1:nvt))
    x%id(1:nid)    = ws%ids(1:nid)
    x%nrep(1:nid)  = ws%nrep(1:nid)
    x%texp(1:nid)  = ws%texp(1:nid)
    x%ivrep(1:nvt) = ws%ivrep(1:nvt)

    isave = imode
    call aderr_set_mode(UW_V_MODE)
    tfoo = fn(a)
    call aderr_set_mode(isave)
    call aderr_grad(der, fn, a)

    x%mean = tfoo%mean
    x%data = 0.0_DP
    do i = 1, x%nid
       call x%get_offset(is_r,ie_r,i,0)
       do j = 1, no
          if (ws%map(i,j).ne.0) then 
             call a(j)%get_offset(is_s,ie_s,ws%map(i,j),0)
             x%data(is_r:ie_r) = x%data(is_r:ie_r)  + &
                  der(j) * a(j)%data(is_s:ie_s)
          end if
       end do
    end do

    return
  end subroutine addobs_f
  
! ********************************
! *
  module subroutine addobs_s(x, a, sn)
! *
! ********************************
    type (uwreal), intent (inout) :: x(:)
    type (uwreal), intent (in)    :: a(:)

    integer        :: no, nid, i, j, is_r, ie_r, &
         is_s, ie_s, nvt, ndo, k, isave
    real (kind=DP) :: der(size(x),size(a))
    type (uwreal)  :: tfoo(size(x))

    interface 
       subroutine sn(r,z)
         USE numtypes
         import :: uwreal
         type (uwreal), intent (inout) :: r(:)
         type (uwreal), intent (in)    :: z(:)
       end subroutine sn
    end interface

    no  = size(a)
    ndo = size(x)
    nvt = 0
    nid = 0
    do i = 1, no
       nvt = nvt + size(a(i)%ivrep)
       nid = nid + a(i)%nid
    end do
    call init_ws(nid,nvt,no)
    call combine_ids_multi(nid,a)
    nvt = sum(ws%nrep(1:nid))

    do i = 1, ndo
       call x(i)%init(nid, ws%nd, ws%nrep, ws%ivrep(1:nvt))
       x(i)%id(1:nid)    = ws%ids(1:nid)
       x(i)%nrep(1:nid)  = ws%nrep(1:nid)
       x(i)%texp(1:nid)  = ws%texp(1:nid)
       x(i)%ivrep(1:nvt) = ws%ivrep(1:nvt)
    end do
    isave = imode
    call aderr_set_mode(UW_V_MODE)
    call sn(tfoo,a)
    call aderr_set_mode(isave)
    call aderr_grad(der, sn, a)

    forall (i=1:ndo)
       x(i)%mean = tfoo(i)%mean
       x(i)%data = 0.0_DP
    end forall
    do k = 1, ndo
       do i = 1, x(k)%nid
          call x(k)%get_offset(is_r,ie_r,i,0)
          do j = 1, no
             if (ws%map(i,j).ne.0) then 
                call a(j)%get_offset(is_s,ie_s,ws%map(i,j),0)
                x(k)%data(is_r:ie_r) = x(k)%data(is_r:ie_r)  + &
                     der(k,j) * a(j)%data(is_s:ie_s)
             end if
          end do
       end do
    end do
       
    return
  end subroutine addobs_s
  
! ********************************
! *
  module subroutine addobs_d(x, a, der, mns)
! *
! ********************************
    type (uwreal), intent  (inout) :: x(:)
    type (uwreal), intent  (in)    :: a(:)
    real (kind=DP), intent (in)    :: der(size(x),size(a)), mns(size(x))

    integer        :: no, nid, i, j, is_r, ie_r, &
         is_s, ie_s, nvt, ndo, k

    no   = size(a)
    ndo  = size(x)
    nvt  = 0
    nid  = 0

    do i = 1, no
       nvt = nvt + size(a(i)%ivrep)
       nid = nid + a(i)%nid
    end do
    call init_ws(nid,nvt,no)
    call combine_ids_multi(nid,a)
    nvt = sum(ws%nrep(1:nid))

    do i = 1, ndo
       call x(i)%init(nid, ws%nd, ws%nrep, ws%ivrep(1:nvt))
       x(i)%id(1:nid)    = ws%ids(1:nid)
       x(i)%nrep(1:nid)  = ws%nrep(1:nid)
       x(i)%texp(1:nid)  = ws%texp(1:nid)
       x(i)%ivrep(1:nvt) = ws%ivrep(1:nvt)
    end do

    forall (i=1:ndo)
       x(i)%mean = mns(i)
       x(i)%data = 0.0_DP
    end forall
    do k = 1, ndo
       do i = 1, x(k)%nid
          call x(k)%get_offset(is_r,ie_r,i,0)
          do j = 1, no
             if (ws%map(i,j).ne.0) then 
                call a(j)%get_offset(is_s,ie_s,ws%map(i,j),0)
                x(k)%data(is_r:ie_r) = x(k)%data(is_r:ie_r)  + &
                     der(k,j) * a(j)%data(is_s:ie_s)
             end if
          end do
       end do
    end do
       
    return
  end subroutine addobs_d
  
! ********************************
! *
  module subroutine addobs_ds(x, a, der, mns)
! *
! ********************************
    type (uwreal), intent  (inout) :: x
    type (uwreal), intent  (in)    :: a(:)
    real (kind=DP), intent (in)    :: der(:), mns

    integer        :: no, nid, i, j, is_r, ie_r, &
         is_s, ie_s, nvt

    no   = size(a)
    nvt  = 0
    nid  = 0

    do i = 1, no
       nvt = nvt + size(a(i)%ivrep)
       nid = nid + a(i)%nid
    end do
    call init_ws(nid,nvt,no)
    call combine_ids_multi(nid,a)
    nvt = sum(ws%nrep(1:nid))

    call x%init(nid, ws%nd, ws%nrep, ws%ivrep(1:nvt))
    x%id(1:nid)    = ws%ids(1:nid)
    x%nrep(1:nid)  = ws%nrep(1:nid)
    x%texp(1:nid)  = ws%texp(1:nid)
    x%ivrep(1:nvt) = ws%ivrep(1:nvt)

    x%mean = mns
    x%data = 0.0_DP
    do i = 1, x%neid()
       call x%get_offset(is_r,ie_r,i,0)
       do j = 1, no
          if (ws%map(i,j).ne.0) then 
             call a(j)%get_offset(is_s,ie_s,ws%map(i,j),0)
             x%data(is_r:ie_r) = x%data(is_r:ie_r)  + &
                  der(j) * a(j)%data(is_s:ie_s)
          end if
       end do
    end do
       
    return
  end subroutine addobs_ds
  
! ********************************
! *
  module subroutine uwdobs_f(x, a, fn)
! *
! ********************************
    type (uwreal), intent (inout) :: x
    type (uwreal), intent (in)    :: a(:)

    integer        :: no, nid, i, j, is_r, ie_r, &
         is_s, ie_s, nvt, ndt
    real (kind=DP) :: der(size(a)), err(size(a))

    interface 
       function fn(z)
         USE numtypes
         real (kind=DP) :: fn
         real (kind=DP), intent (in) :: z(:)
       end function fn
    end interface

    no  = size(a)
    nvt = 0
    nid = 0
    do i = 1, no
       nvt = nvt + size(a(i)%ivrep)
       nid = nid + a(i)%nid
    end do
    call init_ws(nid,nvt,no)
    call combine_ids_multi(nid,a)
    nvt = sum(ws%nrep(1:nid))
    call x%init(nid, ws%nd, ws%nrep, ws%ivrep(1:nvt))
    x%id(1:nid)    = ws%ids(1:nid)
    x%nrep(1:nid)  = ws%nrep(1:nid)
    x%texp(1:nid)  = ws%texp(1:nid)
    x%s(1:nid)     = ws%s(1:nid)
    x%ivrep(1:nvt) = ws%ivrep(1:nvt)
    
    x%mean = fn(a(:)%mean)
    do i = 1, no
       ndt = size(a(i)%data)
       err(i) = sqrt( sum(a(i)%data**2)/real(ndt,kind=DP) ) * err_factor
       if (err(i).gt.a(i)%mean/10.0_DP * err_factor) err(i) = a(i)%mean/10.0_DP * err_factor
    end do
    call richardson(der, fn, a(:)%mean, err)
    
    x%data = 0.0_DP
    do i = 1, x%nid
       call x%get_offset(is_r,ie_r,i,0)
       do j = 1, no
          if (ws%map(i,j).ne.0) then 
             call a(j)%get_offset(is_s,ie_s,ws%map(i,j),0)
             x%data(is_r:ie_r) = x%data(is_r:ie_r)  + &
                  der(j) * a(j)%data(is_s:ie_s)
          end if
       end do
    end do

    return
  end subroutine uwdobs_f
  
! ********************************
! *
  module subroutine uwdobs_s(x, a, sn)
! *
! ********************************
    type (uwreal), intent (inout) :: x(:)
    type (uwreal), intent (in)    :: a(:)

    integer        :: no, nid, i, j, is_r, ie_r, &
         is_s, ie_s, nvt, ndt, ndo, k
    real (kind=DP) :: der(size(x),size(a)), err(size(a))

    interface 
       subroutine sn(r,z)
         USE numtypes
         real (kind=DP), intent (inout) :: r(:)
         real (kind=DP), intent (in)  :: z(:)
       end subroutine sn
    end interface

    no  = size(a)
    ndo = size(x)
    nvt = 0
    nid = 0
    do i = 1, no
       nvt = nvt + size(a(i)%ivrep)
       nid = nid + a(i)%nid
    end do
    call init_ws(nid,nvt,no)
    call combine_ids_multi(nid,a)
    nvt = sum(ws%nrep(1:nid))

    do i = 1, ndo
       call x(i)%init(nid, ws%nd, ws%nrep, ws%ivrep(1:nvt))
       x(i)%id(1:nid)    = ws%ids(1:nid)
       x(i)%nrep(1:nid)  = ws%nrep(1:nid)
       x(i)%texp(1:nid)  = ws%texp(1:nid)
       x(i)%s(1:nid)     = ws%s(1:nid)
       x(i)%ivrep(1:nvt) = ws%ivrep(1:nvt)
    end do
    call sn(x(:)%mean,a(:)%mean)

    do i = 1, no
       ndt = size(a(i)%data)
       err(i) = sqrt( sum(a(i)%data**2)/real(ndt,kind=DP) ) * err_factor
       if (err(i).gt.a(i)%mean/10.0_DP * err_factor) err(i) = a(i)%mean/10.0_DP * err_factor
    end do
    call richardson_multi(der, sn, ndo, a(:)%mean, err)
    
    forall (i=1:ndo) x(i)%data = 0.0_DP
    do k = 1, ndo
       do i = 1, x(k)%nid
          call x(k)%get_offset(is_r,ie_r,i,0)
          do j = 1, no
             if (ws%map(i,j).ne.0) then 
                call a(j)%get_offset(is_s,ie_s,ws%map(i,j),0)
                x(k)%data(is_r:ie_r) = x(k)%data(is_r:ie_r)  + &
                     der(k,j) * a(j)%data(is_s:ie_s)
             end if
          end do
       end do
    end do
       
    return
  end subroutine uwdobs_s
  
! ********************************
! *
  module function uwadd_cr(a,r)
! *
! ********************************
    type (uwreal), intent (in)  :: a
    real (kind=DP), intent (in) :: r
    type (uwreal)               :: uwadd_cr
    
    if (imode.eq.UW_CF_MODE) then
       uwadd_cr = a
       uwadd_cr%mean = a%mean + r
    else
       uwadd_cr = dadd_dr(a,r)
    end if

    return
  end function uwadd_cr
  
! ********************************
! *
  module function uwadd_cl(r,a)
! *
! ********************************
    type (uwreal), intent (in)  :: a
    real (kind=DP), intent (in) :: r
    type (uwreal)               :: uwadd_cl

    if (imode.eq.UW_CF_MODE) then
       uwadd_cl = a
       uwadd_cl%mean = a%mean + r
    else
       uwadd_cl = dadd_dl(r,a)
    end if
       
    return
  end function uwadd_cl
  
! ********************************
! *
  module function uwadd_t(a,b)
! *
! ********************************
    type (uwreal), intent (in) :: a, b
    type (uwreal)              :: uwadd_t

    integer :: i, nid
    integer :: is_r, ie_r, is_s, ie_s, nvt

    if (imode.eq.UW_CF_MODE) then
       nvt = size(a%ivrep) + size(b%ivrep)
       call init_ws(a%nid+b%nid,nvt,2)
       call combine_ids(nid,a,b)
       nvt = sum(ws%nrep(1:nid))
       call uwadd_t%init(nid, ws%nd, ws%nrep, ws%ivrep(1:nvt))
       uwadd_t%id(1:nid)    = ws%ids(1:nid)
       uwadd_t%nrep(1:nid)  = ws%nrep(1:nid)
       uwadd_t%texp(1:nid)  = ws%texp(1:nid)
       uwadd_t%s(1:nid)      = ws%s(1:nid)
       uwadd_t%ivrep(1:nvt) = ws%ivrep(1:nvt)
       
       uwadd_t%mean = a%mean + b%mean
       uwadd_t%data = 0.0_DP
       do i = 1, uwadd_t%nid
          call uwadd_t%get_offset(is_r,ie_r,i,0)
          if (ws%map(i,1).ne.0) then
             call a%get_offset(is_s,ie_s,ws%map(i,1),0)
             uwadd_t%data(is_r:ie_r) = uwadd_t%data(is_r:ie_r) + &
                  a%data(is_s:ie_s)
          end if
          if (ws%map(i,2).ne.0) then
             call b%get_offset(is_s,ie_s,ws%map(i,2),0)
             uwadd_t%data(is_r:ie_r) = uwadd_t%data(is_r:ie_r) + &
                  b%data(is_s:ie_s)
          end if
       end do
    else
       uwadd_t = dadd_dt(a,b)
    end if
       
    return
  end function uwadd_t
  
! ********************************
! *
  module function uwneg(a)
! *
! ********************************
    type (uwreal), intent (in) :: a
    type (uwreal)              :: uwneg

    if (imode.eq.UW_CF_MODE) then
       uwneg = a
       uwneg%mean = -a%mean
       uwneg%data = -a%data
    else
       uwneg = dneg(a)
    end if
       
    return
  end function uwneg
  
! ********************************
! *
  module function uwsub_cr(a,r)
! *
! ********************************
    type (uwreal), intent (in)  :: a
    real (kind=DP), intent (in) :: r
    type (uwreal)               :: uwsub_cr

    if (imode.eq.UW_CF_MODE) then
       uwsub_cr = a
       uwsub_cr%mean = a%mean - r
    else
       uwsub_cr = dsub_dr(a,r)
    end if
       
    return
  end function uwsub_cr
  
! ********************************
! *
  module function uwsub_cl(r,a)
! *
! ********************************
    type (uwreal), intent (in)  :: a
    real (kind=DP), intent (in) :: r
    type (uwreal)               :: uwsub_cl

    if (imode.eq.UW_CF_MODE) then
       uwsub_cl = a
       uwsub_cl%mean = r - a%mean
       uwsub_cl%data = -a%data
    else
       uwsub_cl = dsub_dl(r,a)
    end if
       
    return
  end function uwsub_cl
  
! ********************************
! *
  module function uwsub_t(a,b)
! *
! ********************************
    type (uwreal), intent (in) :: a, b
    type (uwreal)              :: uwsub_t

    integer :: i, nid
    integer :: is_r, ie_r, is_s, ie_s, nvt

    if (imode.eq.UW_CF_MODE) then
       nvt = size(a%ivrep) + size(b%ivrep)
       call init_ws(a%nid+b%nid,nvt,2)
       call combine_ids(nid,a,b)
       nvt = sum(ws%nrep(1:nid))
       call uwsub_t%init(nid, ws%nd, ws%nrep, ws%ivrep(1:nvt))
       uwsub_t%id(1:nid)    = ws%ids(1:nid)
       uwsub_t%nrep(1:nid)  = ws%nrep(1:nid)
       uwsub_t%texp(1:nid)  = ws%texp(1:nid)
       uwsub_t%s(1:nid)     = ws%s(1:nid)
       uwsub_t%ivrep(1:nvt) = ws%ivrep(1:nvt)
       
       uwsub_t%mean = a%mean - b%mean
       uwsub_t%data = 0.0_DP
       do i = 1, uwsub_t%nid
          call uwsub_t%get_offset(is_r,ie_r,i,0)
          if (ws%map(i,1).ne.0) then
             call a%get_offset(is_s,ie_s,ws%map(i,1),0)
             uwsub_t%data(is_r:ie_r) = uwsub_t%data(is_r:ie_r) + &
                  a%data(is_s:ie_s)
          end if
          if (ws%map(i,2).ne.0) then
             call b%get_offset(is_s,ie_s,ws%map(i,2),0)
             uwsub_t%data(is_r:ie_r) = uwsub_t%data(is_r:ie_r) - &
                  b%data(is_s:ie_s)
          end if
       end do
    else
       uwsub_t = dsub_dt(a,b)
    end if
    
    return
  end function uwsub_t
  
! ********************************
! *
  module function uwmul_cr(a,r)
! *
! ********************************
    type (uwreal), intent (in)  :: a
    real (kind=DP), intent (in) :: r
    type (uwreal)               :: uwmul_cr

    if (imode.eq.UW_CF_MODE) then
       uwmul_cr = a
       uwmul_cr%mean = r * a%mean
       uwmul_cr%data = r * uwmul_cr%data
    else
       uwmul_cr = dmul_dr(a,r)
    end if
       
    return
  end function uwmul_cr
  
! ********************************
! *
  module function uwmul_cl(r,a)
! *
! ********************************
    type (uwreal), intent (in)  :: a
    real (kind=DP), intent (in) :: r
    type (uwreal)               :: uwmul_cl

    if (imode.eq.UW_CF_MODE) then
       uwmul_cl = a
       uwmul_cl%mean = r * a%mean
       uwmul_cl%data = r * uwmul_cl%data
    else
       uwmul_cl = dmul_dl(r,a)
    end if
       
    return
  end function uwmul_cl
  
! ********************************
! *
  module function uwmul_t(a,b)
! *
! ********************************
    type (uwreal), intent (in) :: a, b
    type (uwreal)              :: uwmul_t

    integer :: i, nid
    integer :: is_r, ie_r, is_s, ie_s, nvt

    if (imode.eq.UW_CF_MODE) then
       nvt = size(a%ivrep) + size(b%ivrep)
       call init_ws(a%nid+b%nid,nvt,2)
       call combine_ids(nid,a,b)
       nvt = sum(ws%nrep(1:nid))
       call uwmul_t%init(nid, ws%nd, ws%nrep, ws%ivrep(1:nvt))
       uwmul_t%id(1:nid)    = ws%ids(1:nid)
       uwmul_t%nrep(1:nid)  = ws%nrep(1:nid)
       uwmul_t%texp(1:nid)  = ws%texp(1:nid)
       uwmul_t%s(1:nid)     = ws%s(1:nid)
       uwmul_t%ivrep(1:nvt) = ws%ivrep(1:nvt)
       
       uwmul_t%mean = a%mean * b%mean
       uwmul_t%data = 0.0_DP
       do i = 1, uwmul_t%nid
          call uwmul_t%get_offset(is_r,ie_r,i,0)
          if (ws%map(i,1).ne.0) then
             call a%get_offset(is_s,ie_s,ws%map(i,1),0)
             uwmul_t%data(is_r:ie_r) = uwmul_t%data(is_r:ie_r) + &
                  b%mean*a%data(is_s:ie_s)
          end if
          if (ws%map(i,2).ne.0) then
             call b%get_offset(is_s,ie_s,ws%map(i,2),0)
             uwmul_t%data(is_r:ie_r) = uwmul_t%data(is_r:ie_r) + &
                  a%mean*b%data(is_s:ie_s)
          end if
       end do
    else
       uwmul_t = dmul_dt(a,b)
    end if
    
    return
  end function uwmul_t
  
! ********************************
! *
  module function uwdiv_cr(a,r)
! *
! ********************************
    type (uwreal), intent (in)  :: a
    real (kind=DP), intent (in) :: r
    type (uwreal)               :: uwdiv_cr

    if (imode.eq.UW_CF_MODE) then
       uwdiv_cr = a
       uwdiv_cr%mean = a%mean / r
       uwdiv_cr%data = uwdiv_cr%data / r
    else
       uwdiv_cr = ddiv_dr(a,r)
    end if
       
    return
  end function uwdiv_cr
  
! ********************************
! *
  module function uwdiv_cl(r,a)
! *
! ********************************
    type (uwreal), intent (in)  :: a
    real (kind=DP), intent (in) :: r
    type (uwreal)               :: uwdiv_cl

    if (imode.eq.UW_CF_MODE) then
       uwdiv_cl = a
       uwdiv_cl%mean =  r / a%mean
       uwdiv_cl%data = -r / a%mean**2 * a%data
    else
       uwdiv_cl = ddiv_dl(r,a)
    end if
       
    return
  end function uwdiv_cl
  
! ********************************
! *
  module function uwdiv_t(a,b)
! *
! ********************************
    type (uwreal), intent (in) :: a, b
    type (uwreal)              :: uwdiv_t

    integer        :: i, nid
    integer        :: is_r, ie_r, is_s, ie_s, nvt
    real (kind=DP) :: da, db
    
    if (imode.eq.UW_CF_MODE) then
       nvt = size(a%ivrep) + size(b%ivrep)
       call init_ws(a%nid+b%nid,nvt,2)
       call combine_ids(nid,a,b)
       nvt = sum(ws%nrep(1:nid))
       call uwdiv_t%init(nid, ws%nd, ws%nrep, ws%ivrep(1:nvt))
       uwdiv_t%id(1:nid)    = ws%ids(1:nid)
       uwdiv_t%nrep(1:nid)  = ws%nrep(1:nid)
       uwdiv_t%texp(1:nid)  = ws%texp(1:nid)
       uwdiv_t%s(1:nid)     = ws%s(1:nid)
       uwdiv_t%ivrep(1:nvt) = ws%ivrep(1:nvt)
       
       uwdiv_t%mean = a%mean / b%mean
       uwdiv_t%data = 0.0_DP
       da =  1.0_DP/b%mean
       db = -a%mean/b%mean**2
       do i = 1, uwdiv_t%nid
          call uwdiv_t%get_offset(is_r,ie_r,i,0)
          if (ws%map(i,1).ne.0) then
             call a%get_offset(is_s,ie_s,ws%map(i,1),0)
             uwdiv_t%data(is_r:ie_r) = uwdiv_t%data(is_r:ie_r) + &
                  da * a%data(is_s:ie_s)
          end if
          if (ws%map(i,2).ne.0) then
             call b%get_offset(is_s,ie_s,ws%map(i,2),0)
             uwdiv_t%data(is_r:ie_r) = uwdiv_t%data(is_r:ie_r) + &
                  db * b%data(is_s:ie_s)
          end if
       end do
    else
       uwdiv_t = ddiv_dt(a,b)
    end if
    
    return
  end function uwdiv_t
  
! ********************************
! *
  module function uwpow_in(a,n)
! *
! ********************************
    type (uwreal), intent (in)  :: a
    integer, intent (in)        :: n
    type (uwreal) :: uwpow_in

    if (imode.eq.UW_CF_MODE) then
       uwpow_in = a
       uwpow_in%mean = a%mean ** n
       uwpow_in%data = real(n,kind=DP)*a%mean**(n-1) * a%data
    else
       uwpow_in = dpow_in(a,n)
    end if
           
    return
  end function uwpow_in
  
! ********************************
! *
  module function uwpow_re(a,d)
! *
! ********************************
    type (uwreal), intent (in)  :: a
    real (kind=DP), intent (in) :: d
    type (uwreal) :: uwpow_re

    if (imode.eq.UW_CF_MODE) then
       uwpow_re = a
       uwpow_re%mean = a%mean ** d
       uwpow_re%data = d*a%mean**(d-1.0_DP) * a%data
    else
       uwpow_re = dpow_re(a,d)
    end if
       
    return
  end function uwpow_re
  
! ********************************
! *
  module function uwpow_t(a,b)
! *
! ********************************
    type (uwreal), intent (in)  :: a, b
    type (uwreal) :: uwpow_t

    integer :: i, nid
    integer :: is_r, ie_r, is_s, ie_s, nvt
    real (kind=DP) :: da, db
    
    if (imode.eq.UW_CF_MODE) then
       nvt = size(a%ivrep) + size(b%ivrep)
       call init_ws(a%nid+b%nid,nvt,2)
       call combine_ids(nid,a,b)
       nvt = sum(ws%nrep(1:nid))
       call uwpow_t%init(nid, ws%nd, ws%nrep, ws%ivrep(1:nvt))
       uwpow_t%id(1:nid)    = ws%ids(1:nid)
       uwpow_t%nrep(1:nid)  = ws%nrep(1:nid)
       uwpow_t%texp(1:nid)  = ws%texp(1:nid)
       uwpow_t%s(1:nid)     = ws%s(1:nid)
       uwpow_t%ivrep(1:nvt) = ws%ivrep(1:nvt)
       
       uwpow_t%mean = a%mean ** b%mean
       uwpow_t%data = 0.0_DP
       da = uwpow_t%mean*b%mean/a%mean
       db = uwpow_t%mean*log(a%mean)
       do i = 1, uwpow_t%nid
          call uwpow_t%get_offset(is_r,ie_r,i,0)
          if (ws%map(i,1).ne.0) then
             call a%get_offset(is_s,ie_s,ws%map(i,1),0)
             uwpow_t%data(is_r:ie_r) = uwpow_t%data(is_r:ie_r) + &
                  da * a%data(is_s:ie_s)
          end if
          if (ws%map(i,2).ne.0) then
             call b%get_offset(is_s,ie_s,ws%map(i,2),0)
             uwpow_t%data(is_r:ie_r) = uwpow_t%data(is_r:ie_r) + &
                  db * b%data(is_s:ie_s)
          end if
       end do
    else
       uwpow_t = dpow_dt(a,b)
    end if
    
    return
  end function uwpow_t
  
! ********************************
! *
  subroutine init_ws(is,ir,io)
! *
! ********************************
    integer, intent (in) :: is, ir, io
    
    if ((ws%ns.ge.is).and.(ws%no.ge.io).and.(ws%nr.gt.ir)) return
    if (allocated(ws%map))   deallocate(ws%map)
    if (allocated(ws%ids))   deallocate(ws%ids)
    if (allocated(ws%nrep))  deallocate(ws%nrep)
    if (allocated(ws%ivrep)) deallocate(ws%ivrep)
    if (allocated(ws%nd))    deallocate(ws%nd)
    if (allocated(ws%texp))  deallocate(ws%texp)
    if (allocated(ws%s))  deallocate(ws%s)
    
    ws%ns = is
    ws%no = io
    ws%nr = ir
    allocate(ws%map(is,io), ws%ids(is), ws%nrep(is), ws%nd(is), ws%texp(is), ws%s(is))
    allocate(ws%ivrep(ir))
    
    return
  end subroutine init_ws
  
! ********************************
! *
  subroutine combine_ids(nid,a,b)
! *
! ********************************
    type (uwreal), intent (in) :: a, b
    integer, intent (out) :: nid

    integer :: i, j, is, ie, is_r, ie_r, is_a, ie_a
    logical :: fnd, ok
    
    ws%map = 0
    nid = a%nid
    forall (i=1:nid) ws%map(i,1) = i
    ws%ids(1:nid)  = a%id(:)
    ws%nd(1:nid)   = a%nd(1:nid)
    ws%nrep(1:nid) = a%nrep(1:nid)
    ws%texp(1:nid) = a%texp(1:nid)
    ws%s(1:nid)    = a%s(1:nid)
    
    is_r = 1
    ie_r = sum(ws%nrep(1:nid))
    ws%ivrep(is_r:ie_r) = a%ivrep(:)
    is_r = ie_r + 1
    do i = 1, b%nid
       
       fnd = .false.
       do j = 1, a%nid
          if (b%id(i).eq.a%id(j)) then
             ws%map(j,2) = i
             fnd = .true.

             call b%get_vrep_offset(is,ie,i)
             call a%get_vrep_offset(is_a,ie_a,j)
             ok = (ws%nd(j).eq.(b%nd(i)))
             ok = ok.and.((ws%nrep(j).eq.(b%nrep(i))))
             ok = ok.and.(all(ws%ivrep(is_a:ie_a).eq.b%ivrep(is:ie)))

             if (.not.ok) call submodule_error('[combine_ids]', &
                  'data or replica(s) do not match for the same id')

             ok = ( abs(ws%texp(j)-b%texp(i)).lt.1.0E-10_DP )
             if (.not.ok) call submodule_error('[combine_ids]', &
                  'tauexp do not match for the same id')
             exit
          end if
       end do
       
       if (.not.fnd) then
          nid = nid + 1
          ws%ids(nid)   = b%id(i)
          ws%nd(nid)    = b%nd(i)
          ws%nrep(nid)  = b%nrep(i)
          ws%texp(nid)  = b%texp(i)
          ws%s(nid)     = b%s(i)
          
          call b%get_vrep_offset(is,ie,i)
          ie_r = is_r + b%nrep(i) - 1
          ws%ivrep(is_r:ie_r) = b%ivrep(is:ie)
          is_r = ie_r + 1 
          
          ws%map(nid,2) = i
       end if
       
    end do
    
    return
  end subroutine combine_ids

! ********************************
! *
  subroutine combine_ids_multi(nid,a)
! *
! ********************************
    type (uwreal), intent (in) :: a(:)
    integer, intent (out) :: nid

    integer :: i, j, k, is, ie, is_r, ie_r, no
    logical :: fnd, ok
    
    ws%map = 0
    no  = size(a)
    nid = a(1)%nid
    forall (i=1:nid) ws%map(i,1) = i
    ws%ids(1:nid)  = a(1)%id(:)
    ws%nd(1:nid)   = a(1)%nd(1:nid)
    ws%nrep(1:nid) = a(1)%nrep(1:nid)
    ws%texp(1:nid) = a(1)%texp(1:nid)
    ws%s(1:nid) = a(1)%s(1:nid)
    
    is_r = 1
    ie_r = sum(ws%nrep(1:nid))
    ws%ivrep(is_r:ie_r) = a(1)%ivrep(:)
    is_r = ie_r + 1
    do k = 2, no
       do i = 1, a(k)%nid
          
          fnd = .false.
          do j = 1, nid
             if (a(k)%id(i).eq.ws%ids(j)) then
                ws%map(j,k) = i
                fnd = .true.

                ok = (ws%nd(j).eq.(a(k)%nd(i)))
                ok = ok.and.((ws%nrep(j).eq.(a(k)%nrep(i))))
                if (.not.ok) call submodule_error('[combine_ids]', &
                     'data or replica(s) do not match for the same id')

                ok = ( abs(ws%texp(j)-a(k)%texp(i)).lt.1.0E-10_DP )
                if (.not.ok) call submodule_error('[combine_ids]', &
                     'tauexp do not match for the same id')
                exit
             end if
          end do
          
          if (.not.fnd) then
             nid = nid + 1
             ws%ids(nid)   = a(k)%id(i)
             ws%nd(nid)    = a(k)%nd(i)
             ws%nrep(nid)  = a(k)%nrep(i)
             ws%texp(nid)  = a(k)%texp(i)
             ws%s(nid)     = a(k)%s(i)
             
             call a(k)%get_vrep_offset(is,ie,i)
             ie_r = is_r + a(k)%nrep(i) - 1
             ws%ivrep(is_r:ie_r) = a(k)%ivrep(is:ie)
             is_r = ie_r + 1 
             
             ws%map(nid,k) = i
          end if
          
       end do
    end do

    return
  end subroutine combine_ids_multi

! ****************************************
! *
  subroutine richardson(dd, fn, x, hini)
! *
! ****************************************
    real (kind=DP), intent (in)  :: x(:), hini(:)
    real (kind=DP), intent (out) :: dd(size(x))
    
    integer, parameter :: ntab = 20
    real (kind=DP), parameter :: TOL = 1.0, FAC=2.0_DP, &
         FACSQ = 4.0_DP, SAFE=3.0_DP, HMIN=1.0E-6_DP
    integer :: i, j, k, ns
    real (kind=DP) :: h, err, errt, df
    real (kind=DP) :: vt(ntab,ntab), xvp(size(x)), xvm(size(x))

    interface 
       function fn(z)
         USE numtypes
         real (kind=DP) :: fn
         real (kind=DP), intent (in) :: z(:)
       end function fn
    end interface

    ns = size(x)
    do j = 1, ns
       xvp = x
       xvm = x
       
       h = max(hini(j),HMIN)
       xvp(j) = x(j) + h
       xvm(j) = x(j) - h
       vt(1,1) = 0.5_DP*(fn(xvp)-fn(xvm))/h

       err = HUGE(1.0_DP)
       do i = 2, ntab
          h = h/FAC
          xvp(j) = x(j) + h
          xvm(j) = x(j) - h

          vt(1,i) = 0.5_DP*(fn(xvp)-fn(xvm))/h
          df = FACSQ
          do k = 2, i
             vt(k,i) = (vt(k-1,i)*df - vt(k-1,i-1))/(df-1.0_DP)
             df      = df*FACSQ
             errt    = max(abs(vt(k,i)-vt(k-1,i)),abs(vt(k,i)-vt(k-1,i-1)))

             if (errt < err) then
                err   = errt
                dd(j) = vt(k,i)
             endif
          end do

          if (abs(vt(i,i) - vt(i-1,i-1)) > SAFE*err) exit
       end do
    end do
    
    return
  end subroutine richardson

! ****************************************
! *
  subroutine richardson_multi(dd, sn, nf, x, hini)
! *
! ****************************************
    integer, intent (in) :: nf
    real (kind=DP), intent (in)  :: x(:), hini(:)
    real (kind=DP), intent (out) :: dd(nf,size(x))
    
    integer, parameter :: ntab = 20
    real (kind=DP), parameter :: TOL = 1.0, FAC=2.0_DP, &
         FACSQ = 4.0_DP, SAFE=3.0_DP, HMIN=1.0E-6_DP
    integer :: i, j, k, ns, l
    real (kind=DP) :: h, err(nf), errt(nf), df(nf), fvp(nf), fvm(nf)
    real (kind=DP) :: vt(nf,ntab,ntab), xvp(size(x)), xvm(size(x))
    logical :: ddone(nf)
    
    interface 
       subroutine sn(r,z)
         USE numtypes
         real (kind=DP), intent (inout) :: r(:)
         real (kind=DP), intent (in)  :: z(:)
       end subroutine sn
    end interface

    vt = 0.0_DP
    ns = size(x)
    do j = 1, ns
       ddone = .false.
       xvp = x
       xvm = x
       
       h = max(hini(j),HMIN)
       xvp(j) = x(j) + h
       xvm(j) = x(j) - h
       call sn(fvp, xvp)
       call sn(fvm, xvm)
       forall (l=1:nf) vt(l,1,1) = 0.5_DP*(fvp(l)-fvm(l))/h

       err = HUGE(1.0_DP)
       do i = 2, ntab
          h = h/FAC
          xvp(j) = x(j) + h
          xvm(j) = x(j) - h

          call sn(fvp, xvp)
          call sn(fvm, xvm)
          do l = 1, nf
             if (.not.ddone(l)) then
                vt(l,1,i) = 0.5_DP*(fvp(l)-fvm(l))/h
                df(l)     = FACSQ
                do k = 2, i
                   vt(l,k,i) = (vt(l,k-1,i)*df(l) - &
                        vt(l,k-1,i-1))/(df(l)-1.0_DP)
                   
                   df(l)     = df(l)*FACSQ
                   errt(l)   = max(abs(vt(l,k,i)-vt(l,k-1,i)),&
                        abs(vt(l,k,i)-vt(l,k-1,i-1)))
                   
                   if (errt(l) < err(l)) then
                      err(l)    = errt(l)
                      dd(l,j)   = vt(l,k,i)
                   end if
                end do
                ddone(l) = (abs(vt(l,i,i) - vt(l,i-1,i-1)) > SAFE*err(l))
             end if
          end do
             
          if (all(ddone)) exit
       end do
    end do
    
    return
  end subroutine richardson_multi

! ********************************
! *
    subroutine submodule_error(routine, msg)
! *
! ********************************

      Character (len=*), Intent (in) :: routine, msg

      Write(error_unit,*)'error '//Trim(routine)// ':'
      Write(error_unit,'(5X,1A)')Trim(msg)
      Write(error_unit,*)

      stop
    end subroutine submodule_error

  
end submodule aderrors_arith

! Local Variables:
! compile-command: "gmake -C /usr/home/alberto/code/fortran/libana"
! End:

