!
! MODULE uwerrors_cf
!
! "THE BEER-WARE LICENSE":
! Alberto Ramos wrote this file. As long as you retain this 
! notice you can do whatever you want with this stuff. If we meet some 
! day, and you think this stuff is worth it, you can buy me a beer in 
! return. <alberto.ramos@cern.ch>
!
! file:    uwerrors_cf.f90
! created: Thu May 10 17:55:14 2018
!                               

submodule (aderrors) aderrors_cf

  use constants
  use fourier
  
  implicit none

  type ws_alloc
     real (kind=DP), allocatable     :: cf(:), w(:), cdata(:)
     integer, allocatable :: ip(:)
     integer :: nt = -1
  end type ws_alloc
  
  type (ws_alloc), save :: ws
  
contains

! ****************************************
! *
  subroutine init_ws(n1, nf, n)
! *
! ****************************************
    integer, intent (in)  :: n1
    integer, intent (out) :: nf, n
    
    nf = int(log(real(n1+1))/lge2_dp)+2
    n = 2**(nf+1)
    if (n1.gt.ws%nt) then
       if (allocated(ws%cf))    deallocate(ws%cf)
       if (allocated(ws%cdata)) deallocate(ws%cdata)
       if (allocated(ws%ip))    deallocate(ws%ip)
       if (allocated(ws%w))     deallocate(ws%w)

       allocate( &
            ws%cdata(0:n-1), &
            ws%cf(0:n1-1), &
            ws%ip(0:3+int(sqrt(real(n)))), &
            ws%w(0:int(n-1)) &
            )

       ws%nt = n1
    end if

    return
  end subroutine init_ws
    
! ****************************************
! *
  subroutine init_cf(p)
! *
! ****************************************
    class (uwreal), intent (inout) :: p

    integer :: n
    
    n = max(sum(p%tmax),1)
    if ( (allocated(p%gamm)).and.(n.eq.size(p%gamm)).and. &
         (allocated(p%drho)).and.(n.eq.size(p%drho)) ) return

    if (allocated(p%gamm)) deallocate(p%gamm)
    if (allocated(p%drho)) deallocate(p%drho)
    allocate(p%gamm(0:n-1), p%drho(0:n-1))

    return
  end subroutine init_cf

! ****************************************
! *
  module subroutine uwerr(p, ids_texp)
! *
! ****************************************
    class (uwreal), intent (inout) :: p
    integer, intent (in), optional :: ids_texp(:)
    
    integer :: n, k, irst, irnd
    logical :: do_texp(p%nid)

    if (present(ids_texp)) then
       do_texp = .false.
       do n = 1, p%nid
          do k = 1, size(ids_texp)
             if (p%id(n).eq.ids_texp(k)) then
                if (p%texp(n).gt.0.0_DP) then
                   do_texp(n) = .true.
                else
                   do_texp(n) = .false.
                end if
                exit
             end if
          end do
       end do
    else
       where (p%texp.gt.0.0_DP)
          do_texp = .true.
       elsewhere
          do_texp = .false.
       end where
    end if
    
    call init_cf(p)
    call autocorrfunc_obs(p)
    call wopt_texp(p,do_texp)
    call tauint(p)
   
    p%var  = 0.0_DP
    p%derr = 0.0_DP
    do n = 1, p%nid
       if (p%nd(n).eq.1) then
          irst     = 1 + sum(p%nd(1:n-1)) 
          p%var(n) = p%data(irst)**2
       else
          irst = sum(p%tmax(1:n-1))
          irnd = irst + p%iw(n)
          
          if ( (do_texp(n))   .and. &
               (p%gamm(irnd) > 0.0_DP).and. &
               (p%iw(n) > 0)         ) then
             
             p%var(n) = p%gamm(irst)/real(p%nd(n),kind=dp) + &
                  2.0_dp * sum(p%gamm(irst+1:irnd-1))/real(p%nd(n),kind=dp) + &
                  2.0_dp*p%texp(n)*p%gamm(irnd)/real(p%nd(n),kind=dp)
          else
             p%var(n) = p%gamm(irst)/real(p%nd(n),kind=dp) + &
                  2.0_dp * sum(p%gamm(irst+1:irnd))/real(p%nd(n),kind=dp)
          end if
             
          if (p%iw(n).gt.0) p%derr = p%derr + &
               p%var(n)*(real(p%iw(n),kind=DP)-0.5_DP) / real(p%nd(n),kind=DP)
       end if
    end do

    p%err  = sqrt(sum(p%var))
    p%derr = sqrt(p%derr)

    return
  end subroutine uwerr
  
! ****************************************
! *
  subroutine autocorrfunc_obs(p)
! *
! ****************************************
    class (uwreal), intent (inout) :: p

    integer        :: ist, ind, j, i, k, n, irst, irnd, icnd, iadd, nrcnt, irmax
    real (kind=dp) :: cont, dbias

    p%gamm = 0.0_DP
    n = 0
    do n = 1, p%nid
       if (p%nd(n).eq.1) cycle
       call p%get_cf_offset(irst,irnd,n)
       irmax = irnd
       do j = 1, p%nrep(n)
          call p%get_offset(ist,ind,n,j)

          call autocf(p%data(ist:ind))
          irnd = min(irst + (ind-ist)/2, irmax)
          icnd = irnd - irst
          p%gamm(irst:irnd) = p%gamm(irst:irnd) + ws%cf(0:icnd)
       end do
    end do
    
    irst = 0
    do n = 1, p%nid
       if (p%nd(n).eq.1) cycle
       call p%get_vrep_offset(ist,ind,n)
       do j = 0, p%tmax(n)-1
          nrcnt = count(p%ivrep(ist:ind)/2 .ge. j)
          p%gamm(j+irst) = p%gamm(j+irst) / real(p%nd(n) - nrcnt*j,kind=dp)
       end do
       irst = irst + p%tmax(n)
    end do

    call wopt(p)
    irst = 0
    do n = 1, p%nid
       if (p%nd(n).eq.1) cycle
       irnd = irst + p%tmax(n) - 1
       dbias = p%gamm(irst) + 2.0_DP*sum(p%gamm(irst+1:irst+p%iw(n)))
       p%gamm(irst:irnd) = p%gamm(irst:irnd) + dbias/real(p%nd(n),kind=DP)
       irst = irst + p%tmax(n)
    end do
    
    irst = 0
    do n = 1, p%nid
       if (p%nd(n).eq.1) cycle
       irnd = irst + p%tmax(n) - 1
       p%drho(irst:irnd) = 0.0_dp

       if (p%gamm(irst) == 0.0_dp) then
          p%drho(irst:irnd) = 0.0_dp
       else
          iadd = p%iw(n)
          do i = irst, irnd
             ist  = max(1,i-irst-iadd) + irst
             ind  = i + iadd

             do k = ist, ind
                if (k.lt.p%tmax(n)+irst) then
                   cont = -2.0_dp*p%gamm(k)*p%gamm(i)/p%gamm(irst)**2
                else
                   cont = 0.0_DP
                end if
                
                if ((i+k-2*irst) < p%tmax(n)) then
                   cont = cont + p%gamm(i+k-irst)/p%gamm(irst)
                end if
                if (abs(i-k) < p%tmax(n)) then
                   cont = cont + p%gamm(abs(i-k)+irst)/p%gamm(irst)
                end if
                
                p%drho(i) = p%drho(i) + cont**2
             end do
             p%drho(i) = sqrt(p%drho(i)/real(p%nd(n),kind=dp))
          end do
       end if

       irst = irst + p%tmax(n)
    end do

    return
  end subroutine autocorrfunc_obs

! ****************************************
! *
  subroutine autocf(data)
! *
! ****************************************

    real (kind=dp), intent (in) :: data(:)

    integer :: i, nt, nfourier, n
    
    nt = size(data)
    call init_ws(nt, nfourier, n)
    ws%cdata = 0.0_dp
    do i = 0, nt-1
       ws%cdata(2*i) = data(i+1)
    end do

    ws%ip(0) = 0
    call cdft(n, 1, ws%cdata, ws%ip, ws%w)
    do i = 0, n/2-1
       ws%cdata(2*i)   = ws%cdata(2*i)**2 + ws%cdata(2*i+1)**2
       ws%cdata(2*i+1) = 0.0_dp
    end do
    
    call cdft(n, -1, ws%cdata, ws%ip, ws%w)
    ws%cdata(0:n-1) = 2.0_dp*ws%cdata(0:n-1)/real(n,kind=dp)
    
    do i = 0, nt-1
       ws%cf(i) = real(ws%cdata(2*i),kind=DP)
    end do
    
    return
  end subroutine autocf

! ********************************  
! *
  subroutine wopt(p)
! *
! ********************************
    type (uwreal), intent (inout) :: p
    
    integer :: i=0, n, ist, ind
    Real (kind=DP) :: tiw, tau, gw, rw

    do n = 1, p%nid
       if (p%nd(n).eq.1) then
          p%iw(n) = 0
          cycle
       end if
       call p%get_cf_offset(ist,ind,n)
       
       tiw = 0.5_DP
       if (p%gamm(ist)==0.0_DP) then
          p%iw(n) = 1
       else
          do i = ist+1, ind
             rw  = real(i-ist,kind=DP)
             tiw = tiw + p%gamm(i)/p%gamm(ist)
             if (tiw < 0.5_DP) then
                tau =  0.01_DP
                gw  = -1.0_DP
             else
                if (tiw.gt.0.5_DP) then
                   tau = p%s(n)/log((2.0_DP*tiw+1.0_DP)/(2.0_DP*tiw-1.0_DP))
                   gw  = exp(-rw/tau) - tau/sqrt(rw*real(p%nd(n),kind=DP))
                else
                   tau = 1.0_DP
                   gw  = 1.0_DP
                end if
             end if

             if (gw < 0.0_DP) exit
          end do
          p%iw(n) = min(i-ist,ind-ist)
       end if
    end do
    
    return
  end subroutine wopt
  
! ********************************  
! *
  subroutine wopt_texp(p, do_texp)
! *
! ********************************
    type (uwreal), intent (inout) :: p
    logical, intent (in)          :: do_texp(p%nid)
    
    integer :: i=0, n, ist, ind

    do n = 1, p%nid
       if (p%nd(n).eq.1) then
          p%iw(n) = 0
          cycle
       end if
       call p%get_cf_offset(ist,ind,n)

       if (do_texp(n)) then
          p%iw(n) = 0
          if (p%gamm(ist).eq.0.0_DP) cycle
          do i = ist+1, ind
             if (p%gamm(i).lt.0.0_DP)   exit
             if (p%dsig(n)*p%drho(i) > p%gamm(i)/p%gamm(ist)) exit
          end do
          p%iw(n) = i-ist-1
       end if
    end do

    return
  end subroutine wopt_texp
  
! ****************************************
! *
  subroutine tauint(p)
! *
! ****************************************
    type (uwreal), intent (inout) :: p

    integer        :: n, irst, irnd
    real (kind=DP) :: ttex
    
    do n = 1, p%nid
       if (p%nd(n).eq.1) then
          p%ti(n)  = 0.5_DP
          p%dti(n) = 0.0_DP
       else
          irst = sum(p%tmax(1:n-1))
          irnd = irst + p%iw(n)
          ttex = 0.0_DP
          if (p%gamm(irst)>0.0_dp) then
             p%ti(n) = 0.5_dp + sum(p%gamm(irst+1:irnd))/p%gamm(irst) 
             ttex    = p%texp(n)*p%gamm(irnd)/p%gamm(irst)
          else
             p%ti(n)  = 0.5_DP
             p%dti(n) = 0.0_DP
          end if
          p%dti(n) = sqrt( p%ti(n)**2 * &
               real(4*p%iw(n) - 2.0_DP*p%ti(n) + 2,kind=dp)/real(p%nd(n),kind=dp) + p%texp(n)**2*p%drho(irnd)**2)
          p%ti(n) = p%ti(n) + ttex
       end if
    end do

    return
  end subroutine tauint
  
end submodule aderrors_cf


! Local Variables:
! compile-command: "gmake -C /usr/home/alberto/code/fortran/libana"
! End:




