!
! A MODULE 
!                               
! "THE BEER-WARE LICENSE":
! Alberto Ramos wrote this file. As long as you retain this 
! notice you can do whatever you want with this stuff. If we meet some 
! day, and you think this stuff is worth it, you can buy me a beer in 
! return. <alberto.ramos@cern.ch>
!
! $ v1.0 201XXXXX $
!                               

submodule (aderrors) aderrors_cov

  use constants
  use fourier
  
  implicit none

  type wsid_alloc
     integer :: ns = -1, no = -1, nr=-1, nid = 0
     integer, allocatable :: map(:,:), ids(:), nrep(:), &
          ivrep(:), nd(:), tmax(:), iw(:)
     real (kind=DP), allocatable :: texp(:), s(:)
  end type wsid_alloc
  
  type wscf_alloc
     real (kind=DP), allocatable :: cf(:,:,:), w(:), cdata(:), cd(:,:), &
          data(:,:), gamm(:,:,:), drho(:,:)
     integer, allocatable :: ip(:)
     integer :: nt = 0, no = 0, maxrl = 0
  end type wscf_alloc

  real (kind=DP), parameter :: SFT_FAC=2.5_DP
  type (wscf_alloc), save :: wscf 
  type (wsid_alloc), save :: wsid
  
contains

! ********************************
! *
  module subroutine aderr_cov(mat,p)
! *
! ********************************
    type (uwreal),  intent (in)    :: p(:)
    real (kind=DP), intent (inout) :: mat(:,:)

    integer :: n, io1, io2, irst, irnd, is1, is2, ie1, ie2
    
    call autocorrfunc_cov(p)
    call wopt()
    
    mat = 0.0_DP
    do io1 = 1, wsid%no
       do io2 = io1, wsid%no
          do n = 1, wsid%nid
             if (wsid%nd(n).eq.1) then
                if ( (wsid%map(n,io1).ne.0).and.(wsid%map(n,io2).ne.0) ) then
                   call p(io1)%get_offset(is1,ie1,wsid%map(n,io1),0)
                   call p(io2)%get_offset(is2,ie2,wsid%map(n,io2),0)
                   mat(io1,io2) = mat(io1,io2) + &
                        p(io1)%data(is1)*p(io2)%data(is2)
                end if
             else
                irst = sum(wsid%tmax(1:n-1))
                irnd = irst + wsid%iw(n)

                if ( (wsid%texp(n)            > 0.0_DP).and. &
                     (wscf%gamm(irnd,io1,io2) > 0.0_DP).and. &
                     (wsid%iw(n)              > 0)         ) then

                   mat(io1,io2) = mat(io1,io2) + &
                        wscf%gamm(irst,io1,io2)/real(wsid%nd(n),kind=dp) + &
                        2.0_DP * sum(wscf%gamm(irst+1:irnd-1,io1,io2))/real(wsid%nd(n),kind=dp) + &
                        2.0_dp*wsid%texp(n)*wscf%gamm(irnd,io1,io2)/real(wsid%nd(n),kind=dp)
                else
                   mat(io1,io2) = mat(io1,io2) + &
                        wscf%gamm(irst,io1,io2)/real(wsid%nd(n),kind=dp) + &
                        2.0_DP * sum(wscf%gamm(irst+1:irnd,io1,io2))/real(wsid%nd(n),kind=dp)
                end if
             end if
          end do
       end do
    end do

    do io1 = 1, wsid%no
       do io2 = io1+1, wsid%no
          mat(io2,io1) = mat(io1,io2)
       end do
    end do
    
    return
  end subroutine aderr_cov

! ********************************
! *
  subroutine init_wsid(is,ir,io)
! *
! ********************************
    integer, intent (in) :: is, ir, io
    
    if (allocated(wsid%map))   deallocate(wsid%map)
    if (allocated(wsid%ids))   deallocate(wsid%ids)
    if (allocated(wsid%nrep))  deallocate(wsid%nrep)
    if (allocated(wsid%ivrep)) deallocate(wsid%ivrep)
    if (allocated(wsid%nd))    deallocate(wsid%nd)
    if (allocated(wsid%texp))  deallocate(wsid%texp)
    if (allocated(wsid%tmax))  deallocate(wsid%tmax)
    if (allocated(wsid%iw))    deallocate(wsid%iw)
    if (allocated(wsid%s))     deallocate(wsid%s)
    
    wsid%ns = is
    wsid%no = io
    wsid%nr = ir
    allocate(wsid%map(is,io), wsid%ids(is), wsid%nrep(is), wsid%s(is), &
         wsid%nd(is), wsid%texp(is), wsid%tmax(is), wsid%iw(is))
    allocate(wsid%ivrep(ir))
    
    return
  end subroutine init_wsid
  
! ****************************************
! *
  subroutine init_wscf(n1, nobs, maxrl, nf, n)
! *
! ****************************************
    integer, intent (in)  :: n1, nobs, maxrl
    integer, intent (out) :: nf, n
    
    nf = int(log(real(n1+1))/lge2_dp)+2
    n = 2**(nf+1)
    if ((n1.gt.wscf%nt).or.(nobs.gt.wscf%no)) then
       if (allocated(wscf%cf))    deallocate(wscf%cf)
       if (allocated(wscf%cd))    deallocate(wscf%cd)
       if (allocated(wscf%cdata)) deallocate(wscf%cdata)
       if (allocated(wscf%ip))    deallocate(wscf%ip)
       if (allocated(wscf%w))     deallocate(wscf%w)

       allocate( &
            wscf%cdata(0:n-1), &
            wscf%cd(0:n-1,nobs), &
            wscf%cf(0:n1-1,nobs,nobs), &
            wscf%ip(0:3+int(sqrt(real(n)))), &
            wscf%w(0:int(n-1)) &
            )

       wscf%nt = n1
       wscf%no = nobs
    end if

    if ((maxrl.gt.wscf%maxrl).or.(nobs.gt.wscf%no)) then
       if (allocated(wscf%gamm)) deallocate(wscf%gamm)
       if (allocated(wscf%data)) deallocate(wscf%data)
       if (allocated(wscf%drho)) deallocate(wscf%drho)
       allocate(wscf%gamm(0:maxrl/2-1,nobs,nobs), &
            wscf%data(maxrl,nobs), wscf%drho(0:maxrl/2-1,nobs) )
       wscf%maxrl = maxrl
    end if
    
    return
  end subroutine init_wscf
    
! ********************************
! *
  subroutine combine_ids_multi(nid,a)
! *
! ********************************
    type (uwreal), intent (in)  :: a(:)
    integer,       intent (out) :: nid

    integer :: i, j, k, is, ie, is_r, ie_r, no
    logical :: fnd, ok
    
    wsid%map = 0
    no  = size(a)
    nid = a(1)%nid
    forall (i=1:nid) wsid%map(i,1) = i
    wsid%ids(1:nid)  = a(1)%id(:)
    wsid%nd(1:nid)   = a(1)%nd(1:nid)
    wsid%nrep(1:nid) = a(1)%nrep(1:nid)
    wsid%texp(1:nid) = a(1)%texp(1:nid)
    wsid%tmax(1:nid) = a(1)%tmax(1:nid)
    wsid%s(1:nid)    = a(1)%s(1:nid)
    
    is_r = 1
    ie_r = sum(wsid%nrep(1:nid))
    wsid%ivrep(is_r:ie_r) = a(1)%ivrep(:)
    is_r = ie_r + 1
    do k = 2, no
       do i = 1, a(k)%nid
          
          fnd = .false.
          do j = 1, nid
             if (a(k)%id(i).eq.wsid%ids(j)) then
                wsid%map(j,k) = i
                fnd = .true.
                
                ok = (wsid%nd(j).eq.(a(k)%nd(i)))
                ok = ok.and.((wsid%nrep(j).eq.(a(k)%nrep(i))))
                if (.not.ok) call submodule_error('[combine_ids]', &
                     'data or replica(s) do not match for the same id')
                exit
             end if
          end do
          
          if (.not.fnd) then
             nid = nid + 1
             wsid%ids(nid)   = a(k)%id(i)
             wsid%nd(nid)    = a(k)%nd(i)
             wsid%nrep(nid)  = a(k)%nrep(i)
             wsid%texp(nid)  = a(k)%texp(i)
             wsid%tmax(nid)  = a(k)%tmax(i)
             wsid%s(nid)     = a(k)%s(i)
             
             call a(k)%get_vrep_offset(is,ie,i)
             ie_r = is_r + a(k)%nrep(i) - 1
             wsid%ivrep(is_r:ie_r) = a(k)%ivrep(is:ie)
             is_r = ie_r + 1 
             
             wsid%map(nid,k) = i
          end if
          
       end do
    end do

    return
  end subroutine combine_ids_multi

! ********************************  
! *
  subroutine wopt()
! *
! ********************************
    
    integer :: i=0, n, ist, ind, io
    Real (kind=DP) :: tiw, tau, gw, rw

    wsid%iw = 0
    do n = 1, wsid%nid
       if (wsid%nd(n).eq.1) then
          wsid%iw(n) = 0
          cycle
       end if

       ist  = sum(wsid%tmax(1:n-1))
       ind  = ist + wsid%tmax(n) - 1
       do io = 1, wsid%no          
          tiw = 0.5_DP
          if (wsid%texp(n) > 0.0_DP) then
             wsid%iw(n) = 0
             if (wscf%gamm(ist,io,io).gt.0.0_DP) then 
                do i = ist+1, ind
                   if (SFT_FAC*wscf%drho(i,io) > wscf%gamm(i,io,io)/wscf%gamm(ist,io,io)) exit
                end do
             end if
             wsid%iw(n) = max(wsid%iw(n),i-ist)
          else
             if (wscf%gamm(ist,io,io)==0.0_DP) then
                wsid%iw(n) = max(1,wsid%iw(n))
             else
                do i = ist+1, ind
                   rw  = real(i-ist,kind=DP)
                   tiw = tiw + wscf%gamm(i,io,io)/wscf%gamm(ist,io,io)
                   if (tiw < 0.5_DP) then
                      tau = 0.01_DP
                      gw  = -1.0_DP
                   else
                      if (tiw.gt.0.5_DP) then
                         tau = wsid%s(n)/log((2.0_DP*tiw+1.0_DP)/(2.0_DP*tiw-1.0_DP))
                         gw  = exp(-rw/tau) - tau/sqrt(rw*real(wsid%nd(n),kind=DP))
                      else
                         tau = 1.0_DP
                         gw  = 1.0_DP
                      end if
                   end if
                   
                   if (gw < 0.0_DP) exit
                end do
                wsid%iw(n) = max(wsid%iw(n),i-ist)
             end if
          end if
       end do
    end do
    
    return
  end subroutine wopt
  
! ****************************************
! *
  subroutine autocorrfunc_cov(p)
! *
! ****************************************
    class (uwreal), intent (in) :: p(:)

    integer :: ist, ind, j, i, irst, irnd, icnd, nrcnt, no, nid, nvt, &
         idx1, idx2, maxrl, n1, n2, is1, ie1, in, io, io1, io2, iadd, k, n
    real (kind=DP) :: cont

    in = 0
    no  = size(p)
    nvt = 0
    nid = 0
    do i = 1, no
       nvt = nvt + size(p(i)%ivrep)
       nid = nid + p(i)%nid
    end do
    call init_wsid(nid,nvt,no)
    call combine_ids_multi(nid,p)
    wsid%nid = nid

    maxrl = sum(wsid%nd(1:nid))
    call init_wscf(0, no, maxrl, n1, n2)
    if (maxrl.eq.0) return

    wscf%gamm = 0.0_DP
    do i = 1, nid
       if (wsid%nd(i).eq.1) cycle
       irst = sum(wsid%tmax(1:i-1))
       irnd = irst + wsid%tmax(i) - 1
       wscf%data = 0.0_DP
       do n2 = 1, wsid%nrep(i)
          do io = 1, no
             if (wsid%map(i,io).ne.0) then
                call p(io)%get_offset(is1, ie1, wsid%map(i,io), n2)
                in = ie1-is1+1
                wscf%data(1:in,io) = p(io)%data(is1:ie1)
                irnd = irst + min(wsid%tmax(i)-1,(ie1-is1)/2)
             end if
          end do
          
          call covcf(wscf%data(1:in,1:no))
          do io1 = 1, no
             do io2 = 1, no
                idx1 = wsid%map(i,io1)
                idx2 = wsid%map(i,io2)
                if ( (idx1.ne.0).and.(idx2.ne.0) ) then
                   ist = sum(wsid%nrep(1:i-1)) + 1
                   ind = ist + wsid%nrep(i) - 1
                   icnd = irnd - irst
                   do j = 0, icnd
                      nrcnt = count(wsid%ivrep(ist:ind)/2 .ge. j)
                      wscf%gamm(irst+j,io1,io2) = wscf%gamm(irst+j,io1,io2) + &
                           wscf%cf(j,io1,io2) / real(wsid%nd(i) - nrcnt*j,kind=dp)
                   end do
                end if
             end do
          end do
       end do
    end do

    do n = 1, wsid%nid
       do io = 1, no
          irst = sum(wsid%tmax(1:n-1))
          if (wsid%nd(n).eq.1) cycle
          irnd = irst + wsid%tmax(n) - 1
          wscf%drho(irst:irnd,io) = 0.0_dp
          
          if (wscf%gamm(irst,io,io) == 0.0_dp) then
             wscf%drho(irst:irnd,io) = 0.0_dp
          else
             do i = irst, irnd
                iadd = min(100,wsid%tmax(n)/2-1)
                ist  = max(1,i-irst-iadd) + irst
                ind  = ist + iadd
                
                do k = ist, ind
                   cont = - 2.0_dp*wscf%gamm(k,io,io)*wscf%gamm(i,io,io) / &
                        wscf%gamm(irst,io,io)**2
                   
                   if ((i+k-2*irst) < wsid%tmax(n)) then
                      cont = cont + wscf%gamm(i+k-irst,io,io)/wscf%gamm(irst,io,io)
                   end if
                   if (abs(i-k) < wsid%tmax(n)) then
                      cont = cont + wscf%gamm(abs(i-k)+irst,io,io) / &
                           wscf%gamm(irst,io,io)
                   end if
                   
                   wscf%drho(i,io) = wscf%drho(i,io) + cont**2
                end do
                wscf%drho(i,io) = sqrt(wscf%drho(i,io)/real(wsid%nd(n),kind=dp))
             end do
          end if

          irst = irst + wsid%tmax(n)
       end do
    end do


!!$    do io = 1, no
!!$       do i = 0, 10
!!$          write(*,*)i, wscf%drho(i, io), wscf%drho(430+i,io)
!!$       end do
!!$    end do
    
    return
  end subroutine autocorrfunc_cov

! ****************************************
! *
  subroutine covcf(data)
! *
! ****************************************
    real (kind=DP), intent (in) :: data(:,:)
!    real (kind=DP), Intent (out), Allocatable :: C(:,:,:)

    Integer :: I, Nt, Nfourier, N, nobs, io1, io2, j

    nobs = size(data,2)
    nt   = size(data,1)
    call init_wscf(nt,nobs,-2,nfourier,n)

!    nfourier = int(log(real(nt+1))/lge2_dp)+2
!    N = 2*2**nfourier
!    allocate(cdata(0:N-1), cd(0:N-1,nobs))

    wscf%cdata  = 0.0_DP
    wscf%cd     = 0.0_DP
    do j = 1, nobs
       do i = 0, nt-1
          wscf%cd(2*i,j) = data(i+1,j)
       end do
    end  do
    
!    ALLOCATE(Ip(0:3+Int(Sqrt(Real(N)))), W(0:Int(N-1)))
    wscf%ip(0) = 0
    do i = 1, nobs
       CALL CDFT(N, 1, wscf%cd(:,i), wscf%ip, wscf%w)
    end do

!    if (allocated(c)) deallocate(c)
!    allocate(c(0:nt-1,nobs,nobs))
    do io1 = 1, nobs
       do io2 = 1, nobs
          do i = 0, N/2-1
             wscf%cdata(2*i)   = wscf%cd(2*i,io1)*wscf%cd(2*i,io2) + &
                  wscf%cd(2*i+1,io1)*wscf%cd(2*i+1,io2) 
             wscf%cdata(2*i+1) = wscf%cd(2*i+1,io1)*wscf%cd(2*i,io2) - &
                  wscf%cd(2*i,io1)*wscf%cd(2*i+1,io2) 
          end do
          CALL CDFT(N, -1, wscf%cdata, wscf%ip, wscf%w)
          wscf%cdata = 2.0_DP*wscf%cdata/real(N,kind=DP)
          wscf%cf(0,io1,io2) = wscf%cdata(0)
          do i = 1, nt-1
             wscf%cf(i,io1,io2) = ( wscf%cdata(2*i) + wscf%cdata(N-2*i) )*0.5_DP
          end do
       end do
    end do

    return
  end subroutine covCF

! ********************************
! *
    subroutine submodule_error(routine, msg)
! *
! ********************************

      Character (len=*), Intent (in) :: routine, msg

      Write(error_unit,*)'error '//Trim(routine)// ':'
      Write(error_unit,'(5X,1A)')Trim(msg)
      Write(error_unit,*)

      stop
    end subroutine submodule_error
  
end submodule aderrors_cov

! Local Variables:
! compile-command: "gmake -C /usr/home/alberto/code/fortran/libana"
! End:

