!
! MODULE diff 
!
! "THE BEER-WARE LICENSE":
! Alberto Ramos wrote this file. As long as you retain this 
! notice you can do whatever you want with this stuff. If we meet some 
! day, and you think this stuff is worth it, you can buy me a beer in 
! return. <alberto.ramos@cern.ch>
!
! file:    uwerrors_diff.f90
! created: Tue Jul  3 16:59:54 2018
!                               

submodule (aderrors) aderrors_diff

  implicit none
  
contains

! *****************************
! *
  module subroutine uwerr_hess_f(hess, f, x)
! *
! *****************************
    type (uwreal),  intent (in) :: x(:)
    real (kind=DP), intent (inout) :: hess(size(x),size(x))

    type (uwreal) :: h(size(x)), res
    integer :: i, j, n, isave
    
    interface 
       function f(x)
         import :: uwreal
         type (uwreal), intent (in) :: x(:)
         type (uwreal) :: f
       end function f
    end interface
    
    isave = imode
    call aderr_set_mode(UW_D_MODE)
    n = size(x)
    do i = 1, n
       h(i) = (/x(i)%mean, 0.0_DP, 0.0_DP, 0.0_DP/)
    end do

    do i = 1, n
       h(i)%dlt(1) = 1.0_DP
       do j = i, n
          h(j)%dlt(2) = 1.0_DP

          res = f(h)
          hess(i,j) = res%dlt(3)
          if (j.gt.i) hess(j,i) = hess(i,j)

          h(j)%dlt(2) = 0.0_DP
       end do
       h(i)%dlt(1) = 0.0_DP
    end do

    call aderr_set_mode(isave)
    
    return
  end subroutine uwerr_hess_f
  
! *****************************
! *
  module subroutine uwerr_hess_s(hess, s, x)
! *
! *****************************
    type (uwreal),  intent (in) :: x(:)
    real (kind=DP), intent (inout) :: hess(:,:,:)

    type (uwreal) :: h(size(x)), res(size(hess,1))
    integer :: i, j, n, m, k, isave
    
    interface 
       subroutine s(f, x)
         import :: uwreal
         type (uwreal), intent (in)    :: x(:)
         type (uwreal), intent (inout) :: f(:)
       end subroutine s
    end interface
    
    isave = imode
    call aderr_set_mode(UW_D_MODE)
    n = size(x)
    m = size(hess,1)
    do i = 1, n
       h(i) = (/x(i)%mean, 0.0_DP, 0.0_DP, 0.0_DP/)
    end do

    do i = 1, n
       h(i)%dlt(1) = 1.0_DP
       do j = i, n
          h(j)%dlt(2) = 1.0_DP

          call s(res,h)
          do k = 1, m
             hess(k,i,j) = res(k)%dlt(3)
          end do
          if (j.gt.i) forall (k=1:m) hess(k,j,i) = hess(k,i,j)

          h(j)%dlt(2) = 0.0_DP
       end do
       h(i)%dlt(1) = 0.0_DP
    end do

    call aderr_set_mode(isave)
    
    return
  end subroutine uwerr_hess_s
  
! *****************************
! *
  module subroutine uwerr_grad_f(grad, f, x)
! *
! *****************************
    type (uwreal), intent (in)    :: x(:)
    real (kind=DP), intent (inout) :: grad(size(x))

    type (uwreal) :: h(size(x)), res
    integer :: i, n, isave

    interface 
       function f(x)
         import :: uwreal
         type (uwreal), intent (in) :: x(:)
         type (uwreal) :: f
       end function f
    end interface

    isave = imode
    call aderr_set_mode(UW_D_MODE)
    n = size(x)
    do i = 1, n
       h(i) = (/x(i)%mean, 0.0_DP, 0.0_DP, 0.0_DP/)
    end do

    do i = 1, n, 2
       h(i)%dlt(1)   = 1.0_DP
       if (i.lt.n) h(i+1)%dlt(2) = 1.0_DP
       res = f(h)
       grad(i) = res%dlt(1)
       if (i.lt.n) then
          grad(i+1)     = res%dlt(2)
          h(i+1)%dlt(2) = 0.0_DP
       end if
       h(i)%dlt(1) = 0.0_DP
    end do

    call aderr_set_mode(isave)
    
    return
  end subroutine uwerr_grad_f

! *****************************
! *
  module subroutine uwerr_grad_s(grad, s, x)
! *
! *****************************
    type (uwreal),  intent (in)    :: x(:)
    real (kind=DP), intent (inout) :: grad(:,:)

    type (uwreal) :: h(size(x)), res(size(grad,1))
    integer :: i, k, n, m, isave

    interface 
       subroutine s(f, x)
         import :: uwreal
         type (uwreal), intent (in)    :: x(:)
         type (uwreal), intent (inout) :: f(:)
       end subroutine s
    end interface
    
    isave = imode
    call aderr_set_mode(UW_D_MODE)

    n = size(grad,2)
    m = size(grad,1)
    do i = 1, n
       h(i) = (/x(i)%mean, 0.0_DP, 0.0_DP, 0.0_DP/)
    end do

    grad = 0.0_DP
    do i = 1, n, 2
       h(i)%dlt(1) = 1.0_DP
       if (i.lt.n) h(i+1)%dlt(2) = 1.0_DP
       call s(res, h)
       do k = 1, m
          grad(k, i) = res(k)%dlt(1)
       end do
       if (i.lt.n) then
          do k = 1, m
             grad(k,i+1) = res(k)%dlt(2)
          end do
          h(i+1)%dlt(2) = 0.0_DP
       end if
       h(i)%dlt(1) = 0.0_DP
    end do
    
    call aderr_set_mode(isave)

    return
  end subroutine uwerr_grad_s

! ********************************
! *
  subroutine submodule_error(routine, msg)
! *
! ********************************

    character (len=*), intent (in) :: routine, msg

    write(error_unit,*)'error '//trim(routine)// ':'
    write(error_unit,'(5x,1a)')trim(msg)
    write(error_unit,*)

    stop
  end subroutine submodule_error

  
end submodule aderrors_diff


! Local Variables:
! compile-command: "gmake -C /usr/home/alberto/code/fortran/libana"
! End:


