!
! A MODULE 
!                               
! "THE BEER-WARE LICENSE":
! Alberto Ramos wrote this file. As long as you retain this 
! notice you can do whatever you want with this stuff. If we meet some 
! day, and you think this stuff is worth it, you can buy me a beer in 
! return. <alberto.ramos@cern.ch>
!
! $ v1.0 201XXXXX $
!                               

submodule (aderrors) aderrors_func

  use constants
  
  implicit none

contains

! ********************************
! *
  module elemental function uwsqrt(a) result(p)
! *
! ********************************
    type (uwreal), intent (in) :: a
    type (uwreal) :: p

    real (kind=DP) :: d, dd

    if (imode.eq.UW_CF_MODE) then
       call p%init(a%nid,a%nd,a%nrep,a%ivrep)
       p%id    = a%id
       p%texp  = a%texp
       p%s     = a%s
       p%tmax  = a%tmax

       if (allocated(p%gamm)) deallocate(p%gamm)
       if (allocated(p%drho)) deallocate(p%drho)
       p%iw   = 0
       p%ti   = 0.0_DP
       p%dti  = 0.0_DP
       p%var  = 0.0_DP

       p%mean = sqrt(a%mean)
       p%data = a%data/(2.0_DP*p%mean)
    else
       p%mean = sqrt(a%mean)
       if (imode.eq.UW_D_MODE) then
          d        = 0.5_DP/p%mean
          dd       = -0.25_DP/(p%mean**3)

          p%dlt(1) = d  * a%dlt(1)
          p%dlt(2) = d  * a%dlt(2)
          p%dlt(3) = dd * a%dlt(1)*a%dlt(2) + a%dlt(3)*d
       end if
    end if
       
    return
  end function uwsqrt

! ********************************
! *
  module elemental function uwlog(a) result(p)
! *
! ********************************
    type (uwreal), intent (in) :: a
    type (uwreal) :: p

    real (kind=DP) :: d, dd

    if (imode.eq.UW_CF_MODE) then
       call p%init(a%nid,a%nd,a%nrep,a%ivrep)
       p%id    = a%id
       p%texp  = a%texp
       p%s     = a%s
       p%tmax  = a%tmax
       
       if (allocated(p%gamm)) deallocate(p%gamm)
       if (allocated(p%drho)) deallocate(p%drho)
       p%iw   = 0
       p%ti   = 0.0_DP
       p%dti  = 0.0_DP
       p%var  = 0.0_DP
       
       p%mean = log(a%mean)
       p%data = a%data/a%mean
    else
       p%mean = log(a%mean)
       if (imode.eq.UW_D_MODE) then
          d      = 1.0_DP/a%mean
          dd     = -1.0_DP/a%mean**2

          p%dlt(1) = d  * a%dlt(1)
          p%dlt(2) = d  * a%dlt(2)
          p%dlt(3) = dd * a%dlt(1)*a%dlt(2) + a%dlt(3)*d
       end if
    end if
    
    return
  end function uwlog

! ********************************
! *
  module elemental function uwlog10(a) result(p)
! *
! ********************************
    type (uwreal), intent (in) :: a
    type (uwreal) :: p

    real (kind=DP) :: d, dd

    if (imode.eq.UW_CF_MODE) then
       call p%init(a%nid,a%nd,a%nrep,a%ivrep)
       p%id    = a%id
       p%texp  = a%texp
       p%s     = a%s
       p%tmax  = a%tmax
       
       if (allocated(p%gamm)) deallocate(p%gamm)
       if (allocated(p%drho)) deallocate(p%drho)
       p%iw   = 0
       p%ti   = 0.0_DP
       p%dti  = 0.0_DP
       p%var  = 0.0_DP
       
       p%mean = log10(a%mean)
       p%data = a%data/a%mean/LGe10_DP
    else
       p%mean = log10(a%mean)
       if (imode.eq.UW_D_MODE) then
          d      = 1.0_DP/a%mean/LGe10_DP
          dd     = -1.0_DP/a%mean**2/LGe10_DP
          
          p%dlt(1) = d  * a%dlt(1)
          p%dlt(2) = d  * a%dlt(2)
          p%dlt(3) = dd * a%dlt(1)*a%dlt(2) + a%dlt(3)*d
       end if
    end if
       
    return
  end function uwlog10

! ********************************
! *
  module elemental function uwexp(a) result(p)
! *
! ********************************
    type (uwreal), intent (in) :: a
    type (uwreal) :: p

    real (kind=DP) :: d, dd

    if (imode.eq.UW_CF_MODE) then
       call p%init(a%nid,a%nd,a%nrep,a%ivrep)
       p%id    = a%id
       p%texp  = a%texp
       p%s     = a%s
       p%tmax  = a%tmax
       
       if (allocated(p%gamm)) deallocate(p%gamm)
       if (allocated(p%drho)) deallocate(p%drho)
       p%iw   = 0
       p%ti   = 0.0_DP
       p%dti  = 0.0_DP
       p%var  = 0.0_DP
       
       p%mean = exp(a%mean)
       p%data = p%mean*a%data
    else
       p%mean = exp(a%mean)
       if (imode.eq.UW_D_MODE) then
          d      = p%mean
          dd     = p%mean

          p%dlt(1) = d  * a%dlt(1)
          p%dlt(2) = d  * a%dlt(2)
          p%dlt(3) = dd * a%dlt(1)*a%dlt(2) + a%dlt(3)*d
       end if
    end if
       
    return
  end function uwexp

! ********************************
! *
  module elemental function uwsin(a) result(p)
! *
! ********************************
    type (uwreal), intent (in) :: a
    type (uwreal) :: p

    real (kind=DP) :: d, dd

    if (imode.eq.UW_CF_MODE) then
       call p%init(a%nid,a%nd,a%nrep,a%ivrep)
       p%id    = a%id
       p%texp  = a%texp
       p%s     = a%s
       p%tmax  = a%tmax

       if (allocated(p%gamm)) deallocate(p%gamm)
       if (allocated(p%drho)) deallocate(p%drho)
       p%iw   = 0
       p%ti   = 0.0_DP
       p%dti  = 0.0_DP
       p%var  = 0.0_DP

       p%mean = sin(a%mean)
       p%data = cos(a%mean)*a%data
    else
       p%mean = sin(a%mean)
       if (imode.eq.UW_D_MODE) then
          d      = cos(a%mean)
          dd     = -p%mean

          p%dlt(1) = d  * a%dlt(1)
          p%dlt(2) = d  * a%dlt(2)
          p%dlt(3) = dd * a%dlt(1)*a%dlt(2) + a%dlt(3)*d
       end if
    end if
       
    return
  end function uwsin

! ********************************
! *
  module elemental function uwcos(a) result(p)
! *
! ********************************
    type (uwreal), intent (in) :: a
    type (uwreal) :: p

    real (kind=DP) :: d, dd

    if (imode.eq.UW_CF_MODE) then
       call p%init(a%nid,a%nd,a%nrep,a%ivrep)
       p%id    = a%id
       p%texp  = a%texp
       p%s     = a%s
       p%tmax  = a%tmax

       if (allocated(p%gamm)) deallocate(p%gamm)
       if (allocated(p%drho)) deallocate(p%drho)
       p%iw   = 0
       p%ti   = 0.0_DP
       p%dti  = 0.0_DP
       p%var  = 0.0_DP

       p%mean =  cos(a%mean)
       p%data = -sin(a%mean)*a%data
    else
       p%mean =  cos(a%mean)
       if (imode.eq.UW_D_MODE) then
          d      = -sin(a%mean)
          dd     = -p%mean

          p%dlt(1) = d  * a%dlt(1)
          p%dlt(2) = d  * a%dlt(2)
          p%dlt(3) = dd * a%dlt(1)*a%dlt(2) + a%dlt(3)*d
       end if
    end if
       
    return
  end function uwcos

! ********************************
! *
  module elemental function uwtan(a) result(p)
! *
! ********************************
    type (uwreal), intent (in) :: a
    type (uwreal) :: p

    real (kind=DP) :: d, dd

    if (imode.eq.UW_CF_MODE) then
       call p%init(a%nid,a%nd,a%nrep,a%ivrep)
       p%id    = a%id
       p%texp  = a%texp
       p%s     = a%s
       p%tmax  = a%tmax

       if (allocated(p%gamm)) deallocate(p%gamm)
       if (allocated(p%drho)) deallocate(p%drho)
       p%iw   = 0
       p%ti   = 0.0_DP
       p%dti  = 0.0_DP
       p%var  = 0.0_DP

       p%mean = tan(a%mean)
       p%data = (1.0_DP + p%mean**2) * a%data
    else
       p%mean = tan(a%mean)
       if (imode.eq.UW_D_MODE) then
          d      = (1.0_DP + p%mean**2)
          dd     = 2.0_DP*p%mean*d

          p%dlt(1) = d  * a%dlt(1)
          p%dlt(2) = d  * a%dlt(2)
          p%dlt(3) = dd * a%dlt(1)*a%dlt(2) + a%dlt(3)*d
       end if
    end if
           
    return
  end function uwtan

! ********************************
! *
  module elemental function uwasin(a) result(p)
! *
! ********************************
    type (uwreal), intent (in) :: a
    type (uwreal) :: p

    real (kind=DP) :: d, dd

    if (imode.eq.UW_CF_MODE) then
       call p%init(a%nid,a%nd,a%nrep,a%ivrep)
       p%id    = a%id
       p%texp  = a%texp
       p%s     = a%s
       p%tmax  = a%tmax

       if (allocated(p%gamm)) deallocate(p%gamm)
       if (allocated(p%drho)) deallocate(p%drho)
       p%iw   = 0
       p%ti   = 0.0_DP
       p%dti  = 0.0_DP
       p%var  = 0.0_DP

       p%mean = asin(a%mean)
       p%data = a%data/sqrt(1.0_DP-a%mean**2)
    else
       p%mean = asin(a%mean)
       if (imode.eq.UW_D_MODE) then
          d      = 1.0_DP/sqrt(1.0_DP-a%mean**2)
          dd     = 2.0_DP*a%mean*d**2           

          p%dlt(1) = d  * a%dlt(1)
          p%dlt(2) = d  * a%dlt(2)
          p%dlt(3) = dd * a%dlt(1)*a%dlt(2) + a%dlt(3)*d
       end if
    end if
       
    return
  end function uwasin

! ********************************
! *
  module elemental function uwacos(a) result(p)
! *
! ********************************
    type (uwreal), intent (in) :: a
    type (uwreal) :: p

    real (kind=DP) :: d, dd

    if (imode.eq.UW_CF_MODE) then
       call p%init(a%nid,a%nd,a%nrep,a%ivrep)
       p%id    = a%id
       p%texp  = a%texp
       p%s     = a%s
       p%tmax  = a%tmax

       if (allocated(p%gamm)) deallocate(p%gamm)
       if (allocated(p%drho)) deallocate(p%drho)
       p%iw   = 0
       p%ti   = 0.0_DP
       p%dti  = 0.0_DP
       p%var  = 0.0_DP

       p%mean =  acos(a%mean)
       p%data = -a%data/sqrt(1.0_DP-a%mean**2)
    else
       p%mean = acos(a%mean)
       if (imode.eq.UW_D_MODE) then
          d      = -1.0_DP/sqrt(1.0_DP-a%mean**2)
          dd     = -2.0_DP*a%mean*d**2           

          p%dlt(1) = d  * a%dlt(1)
          p%dlt(2) = d  * a%dlt(2)
          p%dlt(3) = dd * a%dlt(1)*a%dlt(2) + a%dlt(3)*d
       end if
    end if
           
    return
  end function uwacos

! ********************************
! *
  module elemental function uwatan(a) result(p)
! *
! ********************************
    type (uwreal), intent (in) :: a
    type (uwreal) :: p

    real (kind=DP) :: d, dd
    
    if (imode.eq.UW_CF_MODE) then
       call p%init(a%nid,a%nd,a%nrep,a%ivrep)
       p%id    = a%id
       p%texp  = a%texp
       p%s     = a%s
       p%tmax  = a%tmax

       if (allocated(p%gamm)) deallocate(p%gamm)
       if (allocated(p%drho)) deallocate(p%drho)
       p%iw   = 0
       p%ti   = 0.0_DP
       p%dti  = 0.0_DP
       p%var  = 0.0_DP

       p%mean = atan(a%mean)
       p%data = a%data/(1.0_DP + p%mean**2)
    else
       p%mean =  atan(a%mean)
       if (imode.eq.UW_D_MODE) then
          d      =  1.0_DP/(1.0_DP + p%mean**2)
          dd     = -2.0_DP*a%mean*d**2           

          p%dlt(1) = d  * a%dlt(1)
          p%dlt(2) = d  * a%dlt(2)
          p%dlt(3) = dd * a%dlt(1)*a%dlt(2) + a%dlt(3)*d
       end if
    end if
       
    return
  end function uwatan

! ********************************
! *
  module elemental function uwsinh(a) result(p)
! *
! ********************************
    type (uwreal), intent (in) :: a
    type (uwreal) :: p

    real (kind=DP) :: d, dd

    if (imode.eq.UW_CF_MODE) then
       call p%init(a%nid,a%nd,a%nrep,a%ivrep)
       p%id    = a%id
       p%texp  = a%texp
       p%s     = a%s
       p%tmax  = a%tmax

       if (allocated(p%gamm)) deallocate(p%gamm)
       if (allocated(p%drho)) deallocate(p%drho)
       p%iw   = 0
       p%ti   = 0.0_DP
       p%dti  = 0.0_DP
       p%var  = 0.0_DP

       p%mean = sinh(a%mean)
       p%data = cosh(a%mean)*a%data
    else
       p%mean = sinh(a%mean)
       if (imode.eq.UW_D_MODE) then
          d      = cosh(a%mean)
          dd     = p%mean

          p%dlt(1) = d  * a%dlt(1)
          p%dlt(2) = d  * a%dlt(2)
          p%dlt(3) = dd * a%dlt(1)*a%dlt(2) + a%dlt(3)*d
       end if
    end if
    
    return
  end function uwsinh

! ********************************
! *
  module elemental function uwcosh(a) result(p)
! *
! ********************************
    type (uwreal), intent (in) :: a
    type (uwreal) :: p

    real (kind=DP) :: d, dd

    if (imode.eq.UW_CF_MODE) then
       call p%init(a%nid,a%nd,a%nrep,a%ivrep)
       p%id    = a%id
       p%texp  = a%texp
       p%s     = a%s
       p%tmax  = a%tmax

       if (allocated(p%gamm)) deallocate(p%gamm)
       if (allocated(p%drho)) deallocate(p%drho)
       p%iw   = 0
       p%ti   = 0.0_DP
       p%dti  = 0.0_DP
       p%var  = 0.0_DP

       p%mean = cosh(a%mean)
       p%data = sinh(a%mean)*a%data
    else
       p%mean = cosh(a%mean)
       if (imode.eq.UW_D_MODE) then
          d      = sinh(a%mean)
          dd     = p%mean

          p%dlt(1) = d  * a%dlt(1)
          p%dlt(2) = d  * a%dlt(2)
          p%dlt(3) = dd * a%dlt(1)*a%dlt(2) + a%dlt(3)*d
       end if
    end if
       
    return
  end function uwcosh

! ********************************
! *
  module elemental function uwtanh(a) result(p)
! *
! ********************************
    type (uwreal), intent (in) :: a
    type (uwreal) :: p

    real (kind=DP) :: d, dd

    if (imode.eq.UW_CF_MODE) then
       call p%init(a%nid,a%nd,a%nrep,a%ivrep)
       p%id    = a%id
       p%texp  = a%texp
       p%s     = a%s
       p%tmax  = a%tmax

       if (allocated(p%gamm)) deallocate(p%gamm)
       if (allocated(p%drho)) deallocate(p%drho)
       p%iw   = 0
       p%ti   = 0.0_DP
       p%dti  = 0.0_DP
       p%var  = 0.0_DP

       p%mean = tanh(a%mean)
       p%data = (1.0_DP - p%mean**2)*a%data
    else
       p%mean =  tanh(a%mean)
       if (imode.eq.UW_D_MODE) then
          d      =  (1.0_DP - p%mean**2)
          dd     = -2.0_DP*p%mean*d

          p%dlt(1) = d  * a%dlt(1)
          p%dlt(2) = d  * a%dlt(2)
          p%dlt(3) = dd * a%dlt(1)*a%dlt(2) + a%dlt(3)*d
       end if
    end if
       
    return
  end function uwtanh

end submodule aderrors_func


! Local Variables:
! compile-command: "gmake -C /usr/home/alberto/code/fortran/libana"
! End:




