!
! A MODULE 
!                               
! "THE BEER-WARE LICENSE":
! Alberto Ramos wrote this file. As long as you retain this 
! notice you can do whatever you want with this stuff. If we meet some 
! day, and you think this stuff is worth it, you can buy me a beer in 
! return. <alberto.ramos@cern.ch>
!
! $ v1.0 20180107 $
!                               

module mod_linked_list

  use iso_fortran_env, only : error_unit, output_unit
  use numtypes
  use constants

  implicit none

  private

  type list_node
     type (list_node), pointer :: next => null(), prev => null()
     class (*), pointer :: data
  end type list_node

  type linked_list
     character (len=:), allocatable :: name
     
     type (list_node), pointer, private :: first   => null()
     type (list_node), pointer, private :: last    => null()
     type (list_node), pointer, private :: current => null()

     integer, private :: nrec = 0

   contains
     procedure :: add     => add_node
     procedure :: show    => show_list
     procedure :: scan    => scan_list
     procedure :: setname => setname_list
     
     procedure :: getdata => getdata_list
  end type linked_list

  interface
     module subroutine add_node(this, data)
       class (linked_list), intent (inout) :: this
       class (*), intent (in) :: data
     end subroutine
     module subroutine show_list(this)
       class (linked_list), intent (inout) :: this
     end subroutine
     module function scan_list(this,nr)
       class (linked_list), intent (inout) :: this
       integer, intent (in), optional :: nr
       logical :: scan_list
     end function
     module subroutine getdata_list(this,dt)
       class (linked_list), intent (inout) :: this
       class (*), intent (inout), pointer  :: dt
     end subroutine
     module subroutine setname_list(this,sn)
       class (linked_list), intent (inout) :: this
       character (len=*),   intent (in)    :: sn
     end subroutine 
  end interface
  
  public :: linked_list

contains

  
! ********************************
! *
  subroutine module_error(routine, msg)
! *
! ********************************

    character (len=*), intent (in) :: routine, msg
    
    write(error_unit,*)'in '//trim(routine)// ' :'
    write(error_unit,'(5x,1a)')trim(msg)
    write(error_unit,*)
    
    stop
  end subroutine module_error
  
end module mod_linked_list

