!
! A MODULE 
!                               
! "THE BEER-WARE LICENSE":
! Alberto Ramos wrote this file. As long as you retain this 
! notice you can do whatever you want with this stuff. If we meet some 
! day, and you think this stuff is worth it, you can buy me a beer in 
! return. <alberto.ramos@cern.ch>
!
! $ v1.0 201XXXXX $
!                               

submodule (mod_linked_list) linked_list_basic

  use iso_fortran_env, only : error_unit, output_unit
  use numtypes
  use constants

  implicit none

contains

! ********************************
! *
  module subroutine setname_list(this,sn)
! *
! ********************************
    class (linked_list), intent (inout) :: this
    character (len=*),   intent (in)    :: sn

    this%name = sn

    return
  end subroutine

! ********************************
! *
  module subroutine getdata_list(this,dt)
! *
! ********************************
    class (linked_list), intent (inout) :: this
    class (*), intent (inout), pointer :: dt

    dt => this%current%data

    return
  end subroutine

! ********************************
! *
  module function scan_list(this,nr)
! *
! ********************************
    class (linked_list), intent (inout) :: this
    integer, intent (in), optional :: nr
    logical :: scan_list
    
    integer :: i
    
    if (present(nr)) then
       if (nr.eq.0) then
          this%current => this%first
          if (.not.associated(this%current)) then
             scan_list = .false.
             return
          end if
          scan_list = .true.
          return
       else if (nr.gt.0) then
          do i = 1, nr
             this%current => this%current%next
             if (.not.associated(this%current)) then
                scan_list = .false.
                return
             end if
          end do

          scan_list = .true.
          return
       else if (nr.lt.0) then
          do i = 1, -nr
             this%current => this%current%prev
             if (.not.associated(this%current)) then
                scan_list = .false.
                return
             end if
          end do

          scan_list = .true.
          return
       end if
    else
       this%current => this%current%next
       if (.not.associated(this%current)) then
          scan_list = .false.
          return
       end if

       scan_list = .true.
       return
    end if
    
    return
  end function
    
! ********************************
! *
  module subroutine show_list(this)
! *
! ********************************
    class (linked_list), intent (inout) :: this
    class (*), pointer :: pt
    integer :: i
    
    if (allocated(this%name)) then
       write(*,*)'Linked list name:  ', trim(this%name)
    else
       write(*,*)'Linked list name:  NO NAME'
    end if
    write(*,*)'Number of records: ', this%nrec

    this%current => this%first
    do i = 1, this%nrec

       pt => this%current%data
       select type (pt)
       type is (real (kind=DP))
          write(*,*)'record: ', i, ' data:', pt
       type is (real (kind=SP))
          write(*,*)'record: ', i, ' data:', pt
       type is (complex (kind=DPC))
          write(*,*)'record: ', i, ' data:', pt
       type is (complex (kind=SPC))
          write(*,*)'record: ', i, ' data:', pt
       type is (integer)
          write(*,*)'record: ', i, ' data:', pt
       class default
             write(*,*)'record: ', i, ' cannot print this data type'
       end select

       this%current => this%current%next
    end do
    
    return
  end subroutine
    
! ********************************
! *
  module subroutine add_node(this, data)
! *
! ********************************
    class (linked_list), intent (inout) :: this
    class (*), intent (in) :: data
    
    if (this%nrec .eq. 0) then
       allocate(this%first)
       this%last => this%first
       this%current => this%first
    else
       allocate(this%last%next)
       this%last%next%prev => this%last
       this%last           => this%last%next
       this%current        => this%last
    end if

    allocate(this%current%data, source=data)

    this%nrec = this%nrec + 1
    
    return
  end subroutine add_node
    
! ********************************
! *
  subroutine submodule_error(routine, msg)
! *
! ********************************

    character (len=*), intent (in) :: routine, msg
    
    write(error_unit,*)'in '//trim(routine)// ' :'
    write(error_unit,'(5x,1a)')trim(msg)
    write(error_unit,*)
    
    stop
  end subroutine submodule_error
  
end submodule linked_list_basic

