!
! PROGRAM simple 
!
! "THE BEER-WARE LICENSE":
! Alberto Ramos wrote this file. As long as you retain this 
! notice you can do whatever you want with this stuff. If we meet some 
! day, and you think this stuff is worth it, you can buy me a beer in 
! return. <alberto.ramos@cern.ch>
!
! file:    simple.f90
! created: Tue Aug 28 15:34:12 2018
!                               

program derived

  use ISO_FORTRAN_ENV, Only : error_unit, output_unit
  use numtypes
  use aderrors
  use simulator
  
  implicit none

  integer, parameter  :: nd = 20000
  type (uwreal)       :: x, y
  real (kind=DP)      :: data_x(nd), err, ti, tau(5), lam(5)


  ! Fill arrays data_x(:) with autocorrelated
  ! data from the module simulator.
  tau = (/1.0_DP, 3.0_DP, 4.0_DP, 5.0_DP, 7.34_DP/)
  lam = (/1.00_DP, 0.87_DP, 1.23_DP, 0.56_DP, 0.87_DP/)

  call gen_series(data_x,  err, ti, tau, lam, 0.3_DP)

  ! Load data_x(:) measurements in variable x. Use
  ! default settings (Stau=4, texp=0, 1 replica)
  x = data_x

  ! Exact, transparent error propagation
  y = sin(x)*cos(x**2)/(1.0_DP+x**4) 
  ! Perform error analysis (optimal window)
  call y%uwerr() 

  ! Print results 
  write(*,'(1A,1I6,1A)')'** Measurements: ', nd, ' **   '
  write(*,100)'  - Gamma-method:  ', y%value(),  " +/- ", y%error(), '( tauint: ', &
       y%taui(1), " +/- ", y%dtaui(1), ')'
  
100 FORMAT((2X,1A,1F8.5,1A,1F7.5,5X,1A,1F0.2,1A,1F7.5,1A))
  
  
  stop
end program derived

