!
! PROGRAM hessian 
!
! "THE BEER-WARE LICENSE":
! Alberto Ramos wrote this file. As long as you retain this 
! notice you can do whatever you want with this stuff. If we meet some 
! day, and you think this stuff is worth it, you can buy me a beer in 
! return. <alberto.ramos@cern.ch>
!
! file:    hessian.f90
! created: Mon Sep  3 01:19:36 2018
!                               

program hessian

  use ISO_FORTRAN_ENV, Only : error_unit, output_unit
  use numtypes
  use constants
  use aderrors
  
  implicit none
  
  type (uwreal)  :: va(4)
  real (kind=DP) :: hess(4,4)
  integer        :: i, j
  
  ! X1 AND X2 AT THE MINIMA
  va(1) = (/ -0.43191853747977416_DP, 0.0_DP/)
  va(2) = (/  4.16107180379317750_DP, 0.0_DP/)

  ! CENTRAL VALUES OF (a,b) (ERRORS NOT NEEDED)
  va(3) = (/ 0.8_DP, 0.0_DP/)
  va(4) = (/ 1.3_DP, 0.0_DP/)

  ! COMPUTE THE HESSIAN
  call aderr_hess(hess, fmulti, va)
  write(*,*)'## HESSIAN ##'
  do i = 1, 4
     write(*,'(100ES15.6)')(hess(i,j), j=1, 4)
  end do
  write(*,*)'##         ##'

contains
  function fmulti(x)
    type (uwreal) :: fmulti
    type (uwreal), intent (in) :: x(:)

    fmulti = x(3)*(x(1)-x(4) + x(3)*x(1)**3)**2 + x(4)*sin(x(2) - x(3) + x(4)**2) &
         + x(1)*x(2)

    return
  end function fmulti
end program hessian

