!
! PROGRAM multi 
!
! "THE BEER-WARE LICENSE":
! Alberto Ramos wrote this file. As long as you retain this 
! notice you can do whatever you want with this stuff. If we meet some 
! day, and you think this stuff is worth it, you can buy me a beer in 
! return. <alberto.ramos@cern.ch>
!
! file:    multi.f90
! created: Sun Sep  2 20:27:29 2018
!                               

program multi

  use ISO_FORTRAN_ENV, Only : error_unit, output_unit
  use numtypes
  use aderrors
  use simulator

  implicit none

  integer, parameter :: nd = 200000
  real (kind=DP)     :: data_x(nd), data_y(nd), err, ti, tau(5), lam(5), texp
  type (uwreal)      :: x, y, z, t 
  integer            :: i
  
  ! Fill array data_x(:) with autocorrelated data
  ! from the module simulator. Save slowest mode in texp
  tau  = (/1.0_DP, 3.0_DP, 4.0_DP, 5.0_DP, 73.4_DP/)
  texp = maxval(tau)
  lam  = (/1.00_DP, 0.87_DP, 1.23_DP, 0.56_DP, 0.087_DP/)
  call gen_series(data_x,  err, ti, tau, lam, 0.3_DP)

  
  ! Fill array data_y(:) with autocorrelated
  ! data from the module simulator.
  tau = (/1.4_DP, 2.4_DP, 1.8_DP, 5.1_DP, 7.14_DP/)
  lam = (/2.00_DP, 0.85_DP, 1.33_DP, 2.56_DP, 1.87_DP/)
  call gen_series(data_y,  err, ti, tau, lam, 0.3_DP)

  ! Load data_x(:) in variable x and set ensemble id to 1.
  ! Set exponential autocorrelation time to add tail 
  x = data_x
  call x%set_id(1)
  call x%set_texp(texp)
  ! Set name of ensemble 1
  i = aderr_new_id("Ensemble X", 1)

  
  ! Load data_y(:) in variable x and set ensemble id to 2
  y = data_y
  call y%set_id(2)
  ! Set name of ensemble 2
  i = aderr_new_id("Ensemble Y", 2)

  ! Observable z is just a variable with error.
  ! Treat as another ensemble with id 3
  z = (/3.23_DP, 0.23_DP/) ! z = 3.23(23)
  call z%set_id(3)
  ! Set name of ensemble 
  i = aderr_new_id("Variable z with errors", 3)

  ! Exact, transparent error propagation.
  t = z*x/y - y/(z*x)

  ! Use Stau=1.75 for ensemble with ID 2
  call t%set_stau(1.75_DP,2)
  ! Error analysis (optimal window, tails, ...)
  call t%uwerr()

  write(*,*)"Result: ", t%value(), " +/- ", t%error()
  do i = 1, t%neid()
     write(*,200)trim(t%error_name(i)), &
          ' contribution to error: ', t%error_src(i)*100., '%'
  end do

200 FORMAT(1A22,1A,5X,1F6.2,1A)
  

  stop
end program multi

